/******************************************************************************/
/*                                                                            */
/* Broadcom BCM4400 Linux Network Driver, Copyright (c) 2000 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/* History:                                                                   */
/******************************************************************************/

#include "b44mm.h"



/******************************************************************************/
/* Local functions. */
/******************************************************************************/

int b44_LM_Abort(struct bm4400_hw * bm_dev);
int b44_LM_QueueRxPackets(struct bm4400_hw * bm_dev);
int b44_LM_SetFlowControl(struct bm4400_hw * bm_dev,
    u32 LocalPhyAd, u32 RemotePhyAd);
static u32 b44_GetPhyAdFlowCntrlSettings(struct bm4400_hw * bm_dev);

static int b44_LM_ResetChip(struct bm4400_hw * bm_dev);
static int b44_LM_DisableChip(struct bm4400_hw * bm_dev);
void b44_LM_ClearStats(struct bm4400_hw *bm_dev);
void b44_LM_WriteCam(struct bm4400_hw *bm_dev, u8 *ea,
    u32 camindex);

u32 b44_LM_getsbaddr(struct bm4400_hw *bm_dev, u32 id,
    u32 coreunit);
void b44_LM_sb_core_disable(struct bm4400_hw *bm_dev);
u32 b44_LM_sb_pci_setup(struct bm4400_hw *bm_dev, u32 cores);
u32 b44_LM_sb_coreunit(struct bm4400_hw *bm_dev);
void b44_LM_sb_core_reset(struct bm4400_hw *bm_dev);
u32 b44_LM_sb_coreid(struct bm4400_hw *bm_dev);
u32 b44_LM_sb_corerev(struct bm4400_hw *bm_dev);
u32 b44_LM_sb_iscoreup(struct bm4400_hw *bm_dev);
#ifdef BCM_WOL
static void b44_LM_ftwrite(struct bm4400_hw *bm_dev, u32 *b,
    u32 nbytes, u32 ftaddr);
#endif

#define BCM4710_PCI_DMA		0x40000000	/* Client Mode PCI memory access space (1 GB) */
#define BCM4710_ENUM		0x18000000	/* Beginning of core enumeration space */

struct sbmap bcm4402[] = {
	{SBID_PCI_DMA,		0,	BCM4710_PCI_DMA},
	{SBID_ENUM,		0,	BCM4710_ENUM},
	{SBID_REG_EMAC,		0,	0x18000000},
	{SBID_REG_CODEC,	0,	0x18001000},
	{SBID_REG_PCI,		0,	0x18002000}
};

#ifdef B44_DEBUG
int b44_reset_count = 0;
#endif

/******************************************************************************/
/* External functions. */
/******************************************************************************/

/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/

int b44_LM_QueueRxPackets(struct bm4400_hw * bm_dev)
{
	struct bm4400_packet * packet;
	u32 pa;
	u32 rxout = bm_dev->rxout;
	u32 ctrl;

	packet = (struct bm4400_packet *) QQ_PopHead(&bm_dev->RxPacketFreeQ.Container);
	while(packet) {

		/* Initialize the receive buffer pointer */
		b44_MM_MapRxDma(bm_dev, packet, &pa);

		*((u32 *) packet->u.Rx.pRxBufferVirt) = 0;

		/* prep the descriptor control value */
		ctrl = packet->u.Rx.RxBufferSize;
		if (rxout == (bm_dev->MaxRxPacketDescCnt - 1))
			ctrl |= CTRL_EOT;

		/* init the rx descriptor */
		bm_dev->pRxDesc[rxout].ctrl = ctrl;
		bm_dev->pRxDesc[rxout].addr = pa + bm_dev->dataoffset;

		bm_dev->RxPacketArr[rxout] = packet;
		rxout = (rxout + 1) & (bm_dev->MaxRxPacketDescCnt - 1);

		packet = (struct bm4400_packet *)
			QQ_PopHead(&bm_dev->RxPacketFreeQ.Container);
	} /* while */

	bm_dev->rxout = rxout;
	wmb();

	REG_WR(bm_dev, dmaregs.rcvptr, rxout * sizeof(dmadd_t));
	return 0;
} /* LM_QueueRxPackets */


/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
static int
b44_LM_EepromReadBlock(struct bm4400_hw * bm_dev,
    u32 offset, u32 *pData, u32 size)
{
	int off, nw;
//	u8 chk8;
	int i;
	u32 *buf;

	off = offset;
	nw = ROUNDUP(size, 4);
	buf = (u32 *) pData;

	/* read the sprom */
	for (i = 0; i < nw; i += 4)
		buf[i/4] = REG_RD_OFFSET(bm_dev, 4096 + off + i);

	return 0;
} /* b44_LM_EepromRead */


/******************************************************************************/
/* Description:                                                               */
/*    This routine initializes default parameters and reads the PCI           */
/*    configurations.                                                         */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_GetAdapterInfo(
struct bm4400_hw * bm_dev)
{
    u32 eprom_dw[32];
    u8 *eprom = (u8 *) eprom_dw;
    int Status;
    u32 val;

    /* Get Device Id and Vendor Id */
    Status = b44_MM_ReadConfig32(bm_dev, PCI_VENDOR_ID_REG, &val);
    if(Status != 0)
    {
        return Status;
    }
    bm_dev->PciVendorId = (u16) val;
    bm_dev->PciDeviceId = (u16) (val >> 16);

    Status = b44_MM_ReadConfig32(bm_dev, PCI_REV_ID_REG, &val);
    if(Status != 0)
    {
        return Status;
    }
    bm_dev->PciRevId = (u8) val;

    /* Get subsystem vendor. */
    Status = b44_MM_ReadConfig32(bm_dev, PCI_SUBSYSTEM_VENDOR_ID_REG, &val);
    if(Status != 0)
    {
        return Status;
    }
    bm_dev->PciSubvendorId = (u16) val;

    /* Get PCI subsystem id. */
    bm_dev->PciSubsystemId = (u16) (val >> 16);

    Status = b44_MM_MapMemBase(bm_dev);
    if(Status != 0)
    {
        return Status;
    }
    /* Initialize the memory view pointer. */
    bm_dev->mmio = (bcmenetregs_t *) bm_dev->pMappedMemBase;

    b44_LM_EepromReadBlock(bm_dev, 0, eprom_dw, sizeof(eprom_dw));
    if (eprom[126] != 1)
        return -EIO;

    bm_dev->PermanentNodeAddress[0] = eprom[79];
    bm_dev->PermanentNodeAddress[1] = eprom[78];
    bm_dev->PermanentNodeAddress[2] = eprom[81];
    bm_dev->PermanentNodeAddress[3] = eprom[80];
    bm_dev->PermanentNodeAddress[4] = eprom[83];
    bm_dev->PermanentNodeAddress[5] = eprom[82];

    memcpy(bm_dev->NodeAddress, bm_dev->PermanentNodeAddress, 6);

    bm_dev->PhyAddr = eprom[90] & 0x1f;
    bm_dev->MdcPort = (eprom[90] >> 14) & 0x1;

    /* Initialize the default values. */
    bm_dev->TxPacketDescCnt = DEFAULT_TX_PACKET_DESC_COUNT;
    bm_dev->RxPacketDescCnt = DEFAULT_RX_PACKET_DESC_COUNT;
    bm_dev->MaxRxPacketDescCnt = DMAMAXRINGSZ / sizeof(dmadd_t);
    bm_dev->rxoffset = 30;
    bm_dev->lazyrxfc = 1;
    bm_dev->lazyrxmult = 0;
    bm_dev->lazytxfc = 0;
    bm_dev->lazytxmult = 0;
    bm_dev->intmask = DEF_INTMASK;
    bm_dev->LinkStatus = LM_STATUS_LINK_DOWN;

#ifdef BCM_WOL
    bm_dev->WakeUpMode = LM_WAKE_UP_MODE_NONE;
#endif

    /* Change driver parameters. */
    Status = b44_MM_GetConfig(bm_dev);
    if(Status != 0)
    {
        return Status;
    }
    return 0;
} /* LM_GetAdapterInfo */


/******************************************************************************/
/* Description:                                                               */
/*    This routine sets up receive/transmit buffer descriptions queues.       */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_InitializeAdapter(struct bm4400_hw * bm_dev)
{
    u32 MemPhy, MemBasePhy;
    u8 *pMemVirt, *pMemBase;
    struct bm4400_packet * packet;
    int Status;
    u32 Size;
    u32 align, j;

    /* Intialize the queues. */
    QQ_InitQueue(&bm_dev->RxPacketReceivedQ.Container, 
        MAX_RX_PACKET_DESC_COUNT);
    QQ_InitQueue(&bm_dev->RxPacketFreeQ.Container,
        MAX_RX_PACKET_DESC_COUNT);

    QQ_InitQueue(&bm_dev->TxPacketFreeQ.Container,MAX_TX_PACKET_DESC_COUNT);
    QQ_InitQueue(&bm_dev->TxPacketXmittedQ.Container,MAX_TX_PACKET_DESC_COUNT);

    /* Allocate memory for packet descriptors. */
    Size = (bm_dev->RxPacketDescCnt + 
        bm_dev->TxPacketDescCnt) * B44_MM_PACKET_DESC_SIZE;
    Status = b44_MM_AllocateMemory(bm_dev, Size, (void * *) &packet);
    if(Status != 0)
    {
        return Status;
    }

    for(j = 0; j < bm_dev->TxPacketDescCnt; j++)
    {
        QQ_PushTail(&bm_dev->TxPacketFreeQ.Container, packet);

        packet = (struct bm4400_packet *) ((u8 *) packet + B44_MM_PACKET_DESC_SIZE);
    } /* for(j.. */

    for(j = 0; j < bm_dev->RxPacketDescCnt; j++)
    {
        /* Receive buffer size. */
        packet->u.Rx.RxBufferSize = 1522 + bm_dev->rxoffset;

        /* Add the descriptor to RxPacketFreeQ. */
        QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);

        packet = (struct bm4400_packet *) ((u8 *) packet + B44_MM_PACKET_DESC_SIZE);
    } /* for */

    /* Initialize the rest of the packet descriptors. */
    Status = b44_MM_InitializeUmPackets(bm_dev);
    if(Status != 0)
    {
        return Status;
    } /* if */

    Size = (bm_dev->MaxRxPacketDescCnt + bm_dev->TxPacketDescCnt) *
        sizeof(dmadd_t) + DMARINGALIGN;

    Status = b44_MM_AllocateSharedMemory(bm_dev, Size, (void *) &pMemBase, &MemBasePhy);
    if(Status != 0)
    {
        return Status;
    }

    MemPhy = (MemBasePhy + (DMARINGALIGN - 1)) & ~(DMARINGALIGN - 1);
    align = MemPhy - MemBasePhy;
    pMemVirt = pMemBase + align;

    bm_dev->pRxDesc = (dmadd_t *) pMemVirt;
    bm_dev->RxDescPhy = MemPhy;

    pMemVirt += bm_dev->MaxRxPacketDescCnt * sizeof(dmadd_t);
    MemPhy += bm_dev->MaxRxPacketDescCnt * sizeof(dmadd_t);

    bm_dev->pTxDesc = (dmadd_t *) pMemVirt;
    bm_dev->TxDescPhy = MemPhy;

    pMemVirt += bm_dev->TxPacketDescCnt * sizeof(dmadd_t);
    MemPhy += bm_dev->MaxRxPacketDescCnt * sizeof(dmadd_t);

    /* Initialize the hardware. */
    Status = b44_LM_ResetAdapter(bm_dev, TRUE);
    if(Status != 0)
    {
        return Status;
    }

    /* We are done with initialization. */
    bm_dev->InitDone = TRUE;

    return 0;
} /* LM_InitializeAdapter */


int
b44_LM_DisableChip(struct bm4400_hw * bm_dev)
{

    /* disable emac */
    REG_WR(bm_dev, enetcontrol, EC_ED);
    SPINWAIT((REG_RD(bm_dev, enetcontrol) & EC_ED), 200);

    REG_WR(bm_dev, dmaregs.xmtcontrol, 0);
    REG_WR(bm_dev, dmaregs.rcvcontrol, 0);
    b44_MM_Wait(10);

    return 0;
}

/******************************************************************************/
/* Description:                                                               */
/*    This function reinitializes the adapter.                                */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_ResetAdapter(struct bm4400_hw * bm_dev, char full)
{

    /* Disable interrupt. */
    b44_LM_DisableInterrupt(bm_dev);

    /* Disable transmit and receive DMA engines.  Abort all pending requests. */
    b44_LM_Abort(bm_dev);

    bm_dev->ShuttingDown = FALSE;

    ASSERT(b44_LM_sb_coreid(bm_dev) == SB_ENET);

    bm_dev->corerev = b44_LM_sb_corerev(bm_dev);

    bm_dev->sbmap = bcm4402;

    bm_dev->coreunit = b44_LM_sb_coreunit(bm_dev);

    /* enable pci interrupts, bursts, and prefetch */
    ASSERT((bm_dev->coreunit == 0) || (bm_dev->coreunit == 1));
    bm_dev->pcirev = b44_LM_sb_pci_setup(bm_dev,
        ((bm_dev->coreunit == 0)? SBIV_ENET0: SBIV_ENET1));

    bm_dev->ddoffset = bm_dev->dataoffset =
        b44_LM_getsbaddr(bm_dev, SBID_PCI_DMA, 0);

    b44_LM_ResetChip(bm_dev);

#if 1
    if (bm_dev->InitDone != TRUE) {
        if (bm_dev->MdcPort == bm_dev->coreunit) {
            b44_LM_ResetPhy(bm_dev);
            b44_LM_SetupPhy(bm_dev);
        }
    }
#endif

    b44_LM_SetMacAddress(bm_dev, bm_dev->NodeAddress);

    /* enable crc32 generation */
    REG_OR(bm_dev, emaccontrol, EMC_CG);

    REG_WR(bm_dev, intrecvlazy, (bm_dev->lazyrxfc << IRL_FC_SHIFT));
    if (bm_dev->lazyrxfc > 1)
    {
        REG_OR(bm_dev, intrecvlazy, (bm_dev->lazyrxmult * bm_dev->lazyrxfc));
    }

    /* enable 802.3x tx flow control (honor received PAUSE frames) */
//    REG_WR(bm_dev, rxconfig, ERC_FE | ERC_UF);

    b44_LM_SetReceiveMask(bm_dev, bm_dev->ReceiveMask);

    /* set max frame lengths - account for possible vlan tag */
    REG_WR(bm_dev, rxmaxlength, MAX_ETHERNET_PACKET_SIZE + 32);
    REG_WR(bm_dev, txmaxlength, MAX_ETHERNET_PACKET_SIZE + 32);

    /* set tx watermark */
    REG_WR(bm_dev, txwatermark, 56);

    if (full)
    {
        /* initialize the tx and rx dma channels */
        /* clear tx descriptor ring */
        memset((void*)bm_dev->pTxDesc, 0, (bm_dev->TxPacketDescCnt *
            sizeof(dmadd_t)));

        REG_WR(bm_dev, dmaregs.xmtcontrol, XC_XE);
        REG_WR(bm_dev, dmaregs.xmtaddr, (bm_dev->TxDescPhy +
            bm_dev->ddoffset));

        /* clear rx descriptor ring */
        memset((void*)bm_dev->pRxDesc, 0, (bm_dev->MaxRxPacketDescCnt *
            sizeof(dmadd_t)));

        REG_WR(bm_dev, dmaregs.rcvcontrol, ((bm_dev->rxoffset <<
            RC_RO_SHIFT) | RC_RE));

        REG_WR(bm_dev, dmaregs.rcvaddr, (bm_dev->RxDescPhy +
            bm_dev->ddoffset));

        /* Queue Rx packet buffers. */
        b44_LM_QueueRxPackets(bm_dev);

        MM_ATOMIC_SET(&bm_dev->SendDescLeft, bm_dev->TxPacketDescCnt - 1);
    }
    else
    {
        REG_WR(bm_dev, dmaregs.rcvcontrol, ((bm_dev->rxoffset <<
            RC_RO_SHIFT) | RC_RE));
    }

    /* turn on the emac */
    REG_OR(bm_dev, enetcontrol, EC_EE);

    return 0;
} /* LM_ResetAdapter */


/******************************************************************************/
/* Description:                                                               */
/*    This routine disables the adapter from generating interrupts.           */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_DisableInterrupt(
    struct bm4400_hw * bm_dev)
{
	REG_WR(bm_dev, intmask, 0);
	(void) REG_RD(bm_dev, intmask);	/* sync readback */
	return 0;
}



/******************************************************************************/
/* Description:                                                               */
/*    This routine enables the adapter to generate interrupts.                */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_EnableInterrupt(
    struct bm4400_hw * bm_dev)
{
	REG_WR(bm_dev, intmask, bm_dev->intmask);
	return 0;
}



/******************************************************************************/
/* Description:                                                               */
/*    This routine puts a packet on the wire if there is a transmit DMA       */
/*    descriptor available; otherwise the packet is queued for later          */
/*    transmission.  If the second argue is NULL, this routine will put       */
/*    the queued packet on the wire if possible.                              */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_SendPacket(struct bm4400_hw * bm_dev, struct bm4400_packet * packet)
{
	u32 fragcount;
	u32 len, txout, ctrl;
	u32 pa;
	int first, next;

	txout = bm_dev->txout;

	bm_dev->TxPacketArr[txout] = packet;
	for(fragcount = 0, first = 1, next = 1; next;
		first = 0, fragcount++) {

		ctrl = 0;
		b44_MM_MapTxDma(bm_dev, packet, &pa, &len, fragcount);
		ctrl = len & CTRL_BC_MASK;

		if (first)
			ctrl |= CTRL_SOF;
		if (fragcount == (packet->u.Tx.FragCount - 1)) {
			ctrl |= CTRL_EOF;
			next = 0;
		}
		if (txout == (bm_dev->TxPacketDescCnt - 1)) {
			ctrl |= CTRL_EOT;
		}
		ctrl |= CTRL_IOC;

		/* init the tx descriptor */
		bm_dev->pTxDesc[txout].ctrl = ctrl;
		bm_dev->pTxDesc[txout].addr = pa + bm_dev->dataoffset;

		txout = (txout + 1) & (bm_dev->TxPacketDescCnt - 1);

	}

	MM_ATOMIC_SUB(&bm_dev->SendDescLeft, packet->u.Tx.FragCount);

	bm_dev->txout = txout;

	wmb();

	REG_WR(bm_dev, dmaregs.xmtptr, (txout * sizeof(dmadd_t)));

	return 0;
}


/******************************************************************************/
/* Description:                                                               */
/*    This routine sets the receive control register according to ReceiveMask */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_SetReceiveMask(struct bm4400_hw * bm_dev, u32 Mask)
{
    u32 ReceiveMask;
    u32 j;
    u32 idx = 0;

    ReceiveMask = Mask;

    if(Mask & LM_ACCEPT_UNICAST)
    {
        Mask &= ~LM_ACCEPT_UNICAST;
    }

    if(Mask & LM_ACCEPT_MULTICAST)
    {
        Mask &= ~LM_ACCEPT_MULTICAST;
    }

    if(Mask & LM_ACCEPT_ALL_MULTICAST)
    {
        Mask &= ~LM_ACCEPT_ALL_MULTICAST;
    }

    if(Mask & LM_ACCEPT_BROADCAST)
    {
        Mask &= ~LM_ACCEPT_BROADCAST;
    }

    if(Mask & LM_PROMISCUOUS_MODE)
    {
        Mask &= ~LM_PROMISCUOUS_MODE;
    }

    /* Make sure all the bits are valid before committing changes. */
    if(Mask)
    {
        return LM_STATUS_FAILURE;
    }

	if (ReceiveMask & LM_PROMISCUOUS_MODE)
		REG_OR(bm_dev, rxconfig, ERC_PE);
	else {
		REG_WR(bm_dev, camcontrol, 0);
		/* our local address */
		b44_LM_WriteCam(bm_dev, bm_dev->NodeAddress, idx++);

		/* allmulti or a list of discrete multicast addresses */
		if (ReceiveMask & LM_ACCEPT_ALL_MULTICAST)
			REG_OR(bm_dev, rxconfig, ERC_AM);
		else if (ReceiveMask & LM_ACCEPT_MULTICAST) {
			for(j = 0; j < bm_dev->McEntryCount; j++) {
				b44_LM_WriteCam(bm_dev, bm_dev->McTable[j],
					idx++);
			}
		}

		/* enable cam */
		REG_OR(bm_dev, camcontrol, CC_CE);
	}

	return 0;
} /* LM_SetReceiveMask */



/******************************************************************************/
/* Description:                                                               */
/*    Disable the interrupt and put the transmitter and receiver engines in   */
/*    an idle state.  Also aborts all pending send requests and receive       */
/*    buffers.                                                                */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_Abort(
struct bm4400_hw * bm_dev)
{
    struct bm4400_packet * packet;
    u32 rxin, txin, txdmask;

    b44_LM_DisableInterrupt(bm_dev);

    b44_LM_DisableChip(bm_dev);

    if (!bm_dev->InitDone)
    {
        return 0;
    }

    txdmask = bm_dev->TxPacketDescCnt - 1;
    for (txin = bm_dev->txin; txin != bm_dev->txout;
        txin = (txin + 1) & txdmask)
    {
        if ((packet = bm_dev->TxPacketArr[txin])) {
            QQ_PushTail(&bm_dev->TxPacketXmittedQ.Container, packet);
            bm_dev->TxPacketArr[txin] = 0;
        }
    }

    if(!bm_dev->ShuttingDown)
    {
        /* Indicate packets to the protocol. */
        b44_MM_IndicateTxPackets(bm_dev);

        /* Indicate received packets to the protocols. */
        b44_MM_IndicateRxPackets(bm_dev);
    }
    else
    {
        /* Move the receive packet descriptors in the ReceivedQ to the */
        /* free queue. */
        for(; ;)
        {
            packet = (struct bm4400_packet *) QQ_PopHead(
                &bm_dev->RxPacketReceivedQ.Container);
            if(packet == NULL)
            {
                break;
            }
            QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
        }
    }

    /* Clean up the Receive desc ring. */

    rxin = bm_dev->rxin;
    while(rxin != bm_dev->rxout) {
        packet = bm_dev->RxPacketArr[rxin];

        QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);

        rxin = (rxin + 1) & (bm_dev->MaxRxPacketDescCnt - 1);
    } /* while */

    bm_dev->rxin = rxin;
    return 0;
} /* LM_Abort */



/******************************************************************************/
/* Description:                                                               */
/*    Disable the interrupt and put the transmitter and receiver engines in   */
/*    an idle state.  Aborts all pending send requests and receive buffers.   */
/*    Also free all the receive buffers.                                      */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_Halt(struct bm4400_hw * bm_dev)
{
    struct bm4400_packet * packet;
    u32 EntryCnt;

    b44_LM_Abort(bm_dev);

    /* Get the number of entries in the queue. */
    EntryCnt = QQ_GetEntryCnt(&bm_dev->RxPacketFreeQ.Container);

    /* Make sure all the packets have been accounted for. */
    for(EntryCnt = 0; EntryCnt < bm_dev->RxPacketDescCnt; EntryCnt++)
    {
        packet = (struct bm4400_packet *) QQ_PopHead(&bm_dev->RxPacketFreeQ.Container);
        if (packet == 0)
            break;

        b44_MM_FreeRxBuffer(bm_dev, packet);

        QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
    }

    b44_LM_ResetChip(bm_dev);

    /* Reprogram the MAC address. */
    b44_LM_SetMacAddress(bm_dev, bm_dev->NodeAddress);

    return 0;
} /* LM_Halt */


static int
b44_LM_ResetChip(struct bm4400_hw *bm_dev)
{
	if (!b44_LM_sb_iscoreup(bm_dev)) {
		b44_LM_sb_pci_setup(bm_dev, 
			((bm_dev->coreunit == 0)? SBIV_ENET0: SBIV_ENET1));	
		/* power on reset: reset the enet core */
		b44_LM_sb_core_reset(bm_dev);

		goto chipinreset;
	}

	/* read counters before resetting the chip */
	if (bm_dev->mibgood)
		b44_LM_StatsUpdate(bm_dev);

	REG_WR(bm_dev, intrecvlazy, 0);

	/* disable emac */
	REG_WR(bm_dev, enetcontrol, EC_ED);
	SPINWAIT((REG_RD(bm_dev, enetcontrol) & EC_ED), 200);

	/* reset the dma engines */
	REG_WR(bm_dev, dmaregs.xmtcontrol, 0);
	bm_dev->txin = bm_dev->txout = 0;

	if (REG_RD(bm_dev, dmaregs.rcvstatus) & RS_RE_MASK) {
		/* wait until channel is idle or stopped */
		SPINWAIT(!(REG_RD(bm_dev, dmaregs.rcvstatus) & RS_RS_IDLE),
			100);
	}

	REG_WR(bm_dev, dmaregs.rcvcontrol, 0);
	bm_dev->rxin = bm_dev->rxout = 0;

	REG_WR(bm_dev, enetcontrol, EC_ES);
	SPINWAIT((REG_RD(bm_dev, enetcontrol) & EC_ES), 200);

	b44_LM_sb_core_reset(bm_dev);

chipinreset:
	b44_LM_ClearStats(bm_dev);

	/*
	 * We want the phy registers to be accessible even when
	 * the driver is "downed" so initialize MDC preamble, frequency,
	 * and whether internal or external phy here.
	 */
		/* 4402 has 62.5Mhz SB clock and internal phy */
	REG_WR(bm_dev, mdiocontrol, 0x8d);

	/* some chips have internal phy, some don't */
	if (!(REG_RD(bm_dev, devcontrol) & DC_IP)) {
		REG_WR(bm_dev, enetcontrol, EC_EP);
	} else if (REG_RD(bm_dev, devcontrol) & DC_ER) {
		REG_AND(bm_dev, devcontrol, ~DC_ER);

		b44_MM_Wait(100);

	}

	/* clear persistent sw intstatus */
	bm_dev->intstatus = 0;
	return 0;
}

void
b44_LM_ClearStats(struct bm4400_hw *bm_dev)
{
	/* must clear mib registers by hand */
	REG_WR(bm_dev, mibcontrol, EMC_RZ);
	(void) REG_RD(bm_dev, mib.tx_good_octets);
	(void) REG_RD(bm_dev, mib.tx_good_pkts);
	(void) REG_RD(bm_dev, mib.tx_octets);
	(void) REG_RD(bm_dev, mib.tx_pkts);
	(void) REG_RD(bm_dev, mib.tx_broadcast_pkts);
	(void) REG_RD(bm_dev, mib.tx_multicast_pkts);
	(void) REG_RD(bm_dev, mib.tx_len_64);
	(void) REG_RD(bm_dev, mib.tx_len_65_to_127);
	(void) REG_RD(bm_dev, mib.tx_len_128_to_255);
	(void) REG_RD(bm_dev, mib.tx_len_256_to_511);
	(void) REG_RD(bm_dev, mib.tx_len_512_to_1023);
	(void) REG_RD(bm_dev, mib.tx_len_1024_to_max);
	(void) REG_RD(bm_dev, mib.tx_jabber_pkts);
	(void) REG_RD(bm_dev, mib.tx_oversize_pkts);
	(void) REG_RD(bm_dev, mib.tx_fragment_pkts);
	(void) REG_RD(bm_dev, mib.tx_underruns);
	(void) REG_RD(bm_dev, mib.tx_total_cols);
	(void) REG_RD(bm_dev, mib.tx_single_cols);
	(void) REG_RD(bm_dev, mib.tx_multiple_cols);
	(void) REG_RD(bm_dev, mib.tx_excessive_cols);
	(void) REG_RD(bm_dev, mib.tx_late_cols);
	(void) REG_RD(bm_dev, mib.tx_defered);
	(void) REG_RD(bm_dev, mib.tx_carrier_lost);
	(void) REG_RD(bm_dev, mib.tx_pause_pkts);
	(void) REG_RD(bm_dev, mib.rx_good_octets);
	(void) REG_RD(bm_dev, mib.rx_good_pkts);
	(void) REG_RD(bm_dev, mib.rx_octets);
	(void) REG_RD(bm_dev, mib.rx_pkts);
	(void) REG_RD(bm_dev, mib.rx_broadcast_pkts);
	(void) REG_RD(bm_dev, mib.rx_multicast_pkts);
	(void) REG_RD(bm_dev, mib.rx_len_64);
	(void) REG_RD(bm_dev, mib.rx_len_65_to_127);
	(void) REG_RD(bm_dev, mib.rx_len_128_to_255);
	(void) REG_RD(bm_dev, mib.rx_len_256_to_511);
	(void) REG_RD(bm_dev, mib.rx_len_512_to_1023);
	(void) REG_RD(bm_dev, mib.rx_len_1024_to_max);
	(void) REG_RD(bm_dev, mib.rx_jabber_pkts);
	(void) REG_RD(bm_dev, mib.rx_oversize_pkts);
	(void) REG_RD(bm_dev, mib.rx_fragment_pkts);
	(void) REG_RD(bm_dev, mib.rx_missed_pkts);
	(void) REG_RD(bm_dev, mib.rx_crc_align_errs);
	(void) REG_RD(bm_dev, mib.rx_undersize);
	(void) REG_RD(bm_dev, mib.rx_crc_errs);
	(void) REG_RD(bm_dev, mib.rx_align_errs);
	(void) REG_RD(bm_dev, mib.rx_symbol_errs);
	(void) REG_RD(bm_dev, mib.rx_pause_pkts);
	(void) REG_RD(bm_dev, mib.rx_nonpause_pkts);
	bm_dev->mibgood = TRUE;
}

void
b44_LM_ServiceRxInterrupt(struct bm4400_hw *bm_dev)
{
	u32 rxin, curr, rxdmask;
	unsigned int len;
	int skiplen = 0;
	bcmenetrxh_t *rxh;
	struct bm4400_packet *packet;

	curr = (REG_RD(bm_dev, dmaregs.rcvstatus) & RS_CD_MASK);
	curr = curr / sizeof(dmadd_t);
	rxdmask = bm_dev->MaxRxPacketDescCnt - 1;
	for (rxin = bm_dev->rxin; rxin != curr; rxin = (rxin + 1) & rxdmask)
	{
		packet = bm_dev->RxPacketArr[rxin];
		if (skiplen > 0) {
			packet->PacketStatus = LM_STATUS_FAILURE;
			skiplen -= packet->u.Rx.RxBufferSize;
			if (skiplen < 0)
				skiplen = 0;
			goto rx_err;
		}
		rxh = (bcmenetrxh_t *) packet->u.Rx.pRxBufferVirt;
		len = MM_SWAP_LE16(rxh->len);
		if (len > (packet->u.Rx.RxBufferSize - bm_dev->rxoffset)) {
			packet->PacketStatus = LM_STATUS_FAILURE;
			skiplen = len - (packet->u.Rx.RxBufferSize -
				bm_dev->rxoffset);
		}
		else {
			int i = 0;

			if (len == 0) {
				while ((len == 0) && (i < 5)) {
					b44_MM_Wait(2);
					len = MM_SWAP_LE16(rxh->len);
					i++;
				}
				if (len == 0) {
					packet->PacketStatus =
						LM_STATUS_FAILURE;
					goto rx_err;
				}
			}
			if (MM_SWAP_LE16(rxh->flags) & RXF_ERRORS) {
				packet->PacketStatus = LM_STATUS_FAILURE;
			}
			else {
				packet->PacketStatus = 0;
			}
			packet->PacketSize = len - 4;
		}
rx_err:
		QQ_PushTail(&bm_dev->RxPacketReceivedQ.Container,
			packet);
		curr = (REG_RD(bm_dev, dmaregs.rcvstatus) & RS_CD_MASK) /
			sizeof(dmadd_t);
	}
	bm_dev->rxin = curr;
}

void
b44_LM_ServiceTxInterrupt(struct bm4400_hw *bm_dev)
{
	u32 txin, curr, txdmask;
	struct bm4400_packet *packet;

	curr = (REG_RD(bm_dev, dmaregs.xmtstatus) & XS_CD_MASK);
	curr = curr / sizeof(dmadd_t);
	txdmask = bm_dev->TxPacketDescCnt - 1;
	for (txin = bm_dev->txin; txin != curr; txin = (txin + 1) & txdmask)
	{
		if ((packet = bm_dev->TxPacketArr[txin])) {
			QQ_PushTail(&bm_dev->TxPacketXmittedQ.Container,
				packet);
			bm_dev->TxPacketArr[txin] = 0;
			MM_ATOMIC_ADD(&bm_dev->SendDescLeft,
				packet->u.Tx.FragCount);
		}
	}
	bm_dev->txin = curr;
}

/******************************************************************************/
/* Description:                                                               */
/*    This is the interrupt event handler routine. It acknowledges all        */
/*    pending interrupts and process all pending events.                      */
/*                                                                            */
/* Return:                                                                    */
/*    0                                                       */
/******************************************************************************/
int
b44_LM_ServiceInterrupts(struct bm4400_hw * bm_dev)
{
	u32 intstatus, intmask;

	while (1) {
		intstatus = REG_RD(bm_dev, intstatus);
		intmask = REG_RD(bm_dev, intmask);

		/* defer unsolicited interrupts */
		intstatus &= intmask;

		/* if not for us */
		if (intstatus == 0)
			return 0;

		REG_WR(bm_dev, intstatus, intstatus);

		if (intstatus & I_RI) {
			b44_LM_ServiceRxInterrupt(bm_dev);
		}

		if (intstatus & (I_XI | I_TO)) {
			b44_LM_ServiceTxInterrupt(bm_dev);
			REG_WR(bm_dev, gptimer, 0);
		}
		if (intstatus & I_ERRORS) {
#ifdef B44_DEBUG
			b44_reset_count++;
#endif
			b44_LM_ResetAdapter(bm_dev, TRUE);
			b44_LM_EnableInterrupt(bm_dev);
		}
		if (!QQ_Empty(&bm_dev->RxPacketReceivedQ.Container)) {
			b44_MM_IndicateRxPackets(bm_dev);
		}
		if (!QQ_Empty(&bm_dev->TxPacketXmittedQ.Container)) {
			b44_MM_IndicateTxPackets(bm_dev);
		}
	}
	return 0;
} /* LM_ServiceInterrupts */



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_MulticastAdd(
struct bm4400_hw * bm_dev,
u8 * pMcAddress) {
    u8 * pEntry;
    u32 j;

    pEntry = bm_dev->McTable[0];
    for(j = 0; j < bm_dev->McEntryCount; j++)
    {
        if(memcmp(pEntry, pMcAddress, 6))
        {
            /* Found a match, increment the instance count. */
            pEntry[LM_MC_INSTANCE_COUNT_INDEX] += 1;

            return 0;
        }

        pEntry += LM_MC_ENTRY_SIZE;
    }
    
    if(bm_dev->McEntryCount >= LM_MAX_MC_TABLE_SIZE)
    {
        return -E2BIG;
    }

    pEntry = bm_dev->McTable[bm_dev->McEntryCount];

    memcpy(pEntry, pMcAddress, 6);
    pEntry[LM_MC_INSTANCE_COUNT_INDEX] = 1;

    bm_dev->McEntryCount++;

    b44_LM_SetReceiveMask(bm_dev, bm_dev->ReceiveMask | LM_ACCEPT_MULTICAST);

    return 0;
} /* b44_LM_MulticastAdd */



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_MulticastDel(struct bm4400_hw * bm_dev, u8 * pMcAddress)
{
    u8 * pEntry;
    u32 j;

    pEntry = bm_dev->McTable[0];
    for(j = 0; j < bm_dev->McEntryCount; j++)
    {
        if(memcmp(pEntry, pMcAddress, 6))
        {
            /* Found a match, decrement the instance count. */
            pEntry[LM_MC_INSTANCE_COUNT_INDEX] -= 1;

            /* No more instance left, remove the address from the table. */
            /* Move the last entry in the table to the delete slot. */
            if(pEntry[LM_MC_INSTANCE_COUNT_INDEX] == 0 &&
                bm_dev->McEntryCount > 1)
            {
		memcpy(pEntry, 
                    bm_dev->McTable[bm_dev->McEntryCount-1], 6);
                pEntry[LM_MC_INSTANCE_COUNT_INDEX] =
                    bm_dev->McTable[bm_dev->McEntryCount-1]
                    [LM_MC_INSTANCE_COUNT_INDEX];
            }
            bm_dev->McEntryCount--;

            /* Update the receive mask if the table is empty. */
            if(bm_dev->McEntryCount == 0)
            {
                b44_LM_SetReceiveMask(bm_dev, 
                    bm_dev->ReceiveMask & ~LM_ACCEPT_MULTICAST);
            }

            return 0;
        }

        pEntry += LM_MC_ENTRY_SIZE;
    }

    return -ENOENT;
} /* b44_LM_MulticastDel */



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_MulticastClear(
struct bm4400_hw * bm_dev) {
    bm_dev->McEntryCount = 0;

    b44_LM_SetReceiveMask(bm_dev, bm_dev->ReceiveMask & ~LM_ACCEPT_MULTICAST);

    return 0;
} /* b44_LM_MulticastClear */



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_SetMacAddress(struct bm4400_hw * bm_dev, u8 * pMacAddress)
{
    return 0;
}


/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_SetFlowControl(
    struct bm4400_hw * bm_dev,
    u32 LocalPhyAd,
    u32 RemotePhyAd)
{
    u32 FlowCap;

    /* Resolve flow control. */
    FlowCap = LM_FLOW_CONTROL_NONE;

    /* See Table 28B-3 of 802.3ab-1999 spec. */
    if(bm_dev->FlowControlCap & LM_FLOW_CONTROL_AUTO_PAUSE)
    {
        if(LocalPhyAd & PHY_AN_AD_PAUSE_CAPABLE)
        {
            if(LocalPhyAd & PHY_AN_AD_ASYM_PAUSE)
            {
                if(RemotePhyAd & PHY_LINK_PARTNER_PAUSE_CAPABLE)
                {
                    FlowCap = LM_FLOW_CONTROL_TRANSMIT_PAUSE |
                        LM_FLOW_CONTROL_RECEIVE_PAUSE;
                }
                else if(RemotePhyAd & PHY_LINK_PARTNER_ASYM_PAUSE)
                {
                    FlowCap = LM_FLOW_CONTROL_RECEIVE_PAUSE;
                }
            }
            else
            {
                if(RemotePhyAd & PHY_LINK_PARTNER_PAUSE_CAPABLE)
                {
                    FlowCap = LM_FLOW_CONTROL_TRANSMIT_PAUSE |
                        LM_FLOW_CONTROL_RECEIVE_PAUSE;
                }
            }
        }
        else if(LocalPhyAd & PHY_AN_AD_ASYM_PAUSE)
        {
            if((RemotePhyAd & PHY_LINK_PARTNER_PAUSE_CAPABLE) &&
                (RemotePhyAd & PHY_LINK_PARTNER_ASYM_PAUSE))
            {
                FlowCap = LM_FLOW_CONTROL_TRANSMIT_PAUSE;
            }
        }
    }
    else
    {
        FlowCap = bm_dev->FlowControlCap;
    }

    /* Enable/disable rx PAUSE. */
    if(FlowCap & LM_FLOW_CONTROL_RECEIVE_PAUSE &&
        (bm_dev->FlowControlCap == LM_FLOW_CONTROL_AUTO_PAUSE ||
        bm_dev->FlowControlCap & LM_FLOW_CONTROL_RECEIVE_PAUSE))
    {
        bm_dev->FlowControl |= LM_FLOW_CONTROL_RECEIVE_PAUSE;
        REG_WR(bm_dev, rxconfig, ERC_EF);

    }

    /* Enable/disable tx PAUSE. */
    if(FlowCap & LM_FLOW_CONTROL_TRANSMIT_PAUSE &&
        (bm_dev->FlowControlCap == LM_FLOW_CONTROL_AUTO_PAUSE ||
        bm_dev->FlowControlCap & LM_FLOW_CONTROL_TRANSMIT_PAUSE))
    {
        bm_dev->FlowControl |= LM_FLOW_CONTROL_TRANSMIT_PAUSE;
        REG_WR(bm_dev, emacflowcontrol, EMF_PG | (0xc0 & EMF_RFH_MASK));

    }

    return 0;
}



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
int
b44_LM_SetupPhy(
    struct bm4400_hw * bm_dev)
{
	u32 val;
	u32 Adv, FCAdv, Ctrl;
	int RestartAuto = 0;

	/* enable activity led */
	b44_LM_ReadPhy(bm_dev, 26, &val);
	b44_LM_WritePhy(bm_dev, 26, val & 0x7fff);

	/* enable traffic meter led mode */
	b44_LM_ReadPhy(bm_dev, 27, &val);
	b44_LM_WritePhy(bm_dev, 27, val | (1 << 6));
	if (!bm_dev->DisableAutoNeg) {
		if (bm_dev->RequestedLineSpeed == LM_LINE_SPEED_AUTO) {
			Adv = PHY_AN_AD_ALL_SPEEDS;
		}
		else if (bm_dev->RequestedLineSpeed == LM_LINE_SPEED_10MBPS) {
			if (bm_dev->RequestedDuplexMode ==
				LM_DUPLEX_MODE_FULL) {
				Adv = PHY_AN_AD_10BASET_FULL;
			}
			else {
				Adv = PHY_AN_AD_10BASET_HALF;
			}
		}
		else if (bm_dev->RequestedLineSpeed == LM_LINE_SPEED_100MBPS) {
			if (bm_dev->RequestedDuplexMode ==
				LM_DUPLEX_MODE_FULL) {
				Adv = PHY_AN_AD_100BASETX_FULL;
			}
			else {
				Adv = PHY_AN_AD_100BASETX_HALF;
			}
		}
		else {
			Adv = PHY_AN_AD_ALL_SPEEDS;
		}

		if ((bm_dev->RequestedLineSpeed == LM_LINE_SPEED_AUTO) ||
			(bm_dev->RequestedDuplexMode == LM_DUPLEX_MODE_FULL)) {
			FCAdv = b44_GetPhyAdFlowCntrlSettings(bm_dev);
			val &= PHY_AN_AD_ASYM_PAUSE |
				PHY_AN_AD_PAUSE_CAPABLE;
			if (FCAdv != val) {
				RestartAuto = 1;
				Adv |= FCAdv;
				goto restart_auto_neg;
			}
		}

		b44_LM_ReadPhy(bm_dev, PHY_CTRL_REG, &Ctrl);
		if (!(Ctrl & PHY_CTRL_AUTO_NEG_ENABLE)) {
			RestartAuto = 1;
			goto restart_auto_neg;
		}
		b44_LM_ReadPhy(bm_dev, PHY_AN_AD_REG, &val);
		if ((val & PHY_AN_AD_ALL_SPEEDS) != Adv) {
			RestartAuto = 1;
		}
restart_auto_neg:
		if (RestartAuto) {
			Adv |= PHY_AN_AD_PROTOCOL_802_3_CSMA_CD;
			b44_LM_WritePhy(bm_dev, PHY_AN_AD_REG, Adv);
			b44_LM_WritePhy(bm_dev, PHY_CTRL_REG,
				PHY_CTRL_AUTO_NEG_ENABLE |
				PHY_CTRL_RESTART_AUTO_NEG);
		}
		bm_dev->Advertising = Adv;
	}
	else {
		b44_LM_ReadPhy(bm_dev, PHY_CTRL_REG, &Ctrl);
		Ctrl &= ~(PHY_CTRL_SPEED_SELECT_100MBPS |
			PHY_CTRL_FULL_DUPLEX_MODE | PHY_CTRL_AUTO_NEG_ENABLE);
		if (bm_dev->RequestedLineSpeed == LM_LINE_SPEED_100MBPS) {
			Ctrl |= PHY_CTRL_SPEED_SELECT_100MBPS;
		}
		if (bm_dev->RequestedDuplexMode == LM_DUPLEX_MODE_FULL) {
			Ctrl |= PHY_CTRL_FULL_DUPLEX_MODE;
		}
		b44_LM_WritePhy(bm_dev, PHY_CTRL_REG, Ctrl);
		if (bm_dev->RequestedDuplexMode == LM_DUPLEX_MODE_FULL) {
			bm_dev->FlowControlCap &= ~LM_FLOW_CONTROL_AUTO_PAUSE;
			b44_LM_SetFlowControl(bm_dev, 0, 0);
		}
	}
	return 0;
}

int
b44_LM_ResetPhy(struct bm4400_hw *bm_dev)
{
	u32 value32;

	b44_LM_WritePhy(bm_dev, 0, PHY_CTRL_PHY_RESET);
	b44_MM_Wait(100);
	b44_LM_ReadPhy(bm_dev, 0, &value32);
	if (value32 & PHY_CTRL_PHY_RESET) {
		printf("Phy reset not complete\n");
	}
	return 0;
}

/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
void
b44_LM_ReadPhy(
struct bm4400_hw * bm_dev, u32 PhyReg, u32 *pData32)
{
	/* clear mii_int */
	REG_WR(bm_dev, emacintstatus, EI_MII);

	/* issue the read */
	REG_WR(bm_dev, mdiodata, (MD_SB_START | MD_OP_READ |
		(bm_dev->PhyAddr << MD_PMD_SHIFT)
		| (PhyReg << MD_RA_SHIFT) | MD_TA_VALID));

	/* wait for it to complete */
	SPINWAIT(((REG_RD(bm_dev, emacintstatus) & EI_MII) == 0), 100);
	if ((REG_RD(bm_dev, emacintstatus) & EI_MII) == 0) {
		printf("LM_ReadPhy: did not complete\n");
	}

	*pData32 = REG_RD(bm_dev, mdiodata) & MD_DATA_MASK;
} /* LM_ReadPhy */



/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
void
b44_LM_WritePhy(
struct bm4400_hw * bm_dev, u32 PhyReg, u32 Data32)
{
	/* clear mii_int */
	REG_WR(bm_dev, emacintstatus, EI_MII);
	ASSERT((REG_RD(bm_dev, emacintstatus) & EI_MII) == 0);

	/* issue the write */
	REG_WR(bm_dev, mdiodata, (MD_SB_START | MD_OP_WRITE |
		(bm_dev->PhyAddr << MD_PMD_SHIFT)
		| (PhyReg << MD_RA_SHIFT) | MD_TA_VALID | Data32));

	/* wait for it to complete */
	SPINWAIT(((REG_RD(bm_dev, emacintstatus) & EI_MII) == 0), 100);
	if ((REG_RD(bm_dev, emacintstatus) & EI_MII) == 0) {
		printf("b44_LM_WritePhy: did not complete\n");
	}
} /* LM_WritePhy */


int
b44_LM_StatsUpdate(struct bm4400_hw *bm_dev)
{
	bm_dev->tx_good_octets += REG_RD(bm_dev, mib.tx_good_octets);
	bm_dev->tx_good_pkts += REG_RD(bm_dev, mib.tx_good_pkts);
	bm_dev->tx_octets += REG_RD(bm_dev, mib.tx_octets);
	bm_dev->tx_pkts += REG_RD(bm_dev, mib.tx_pkts);
	bm_dev->tx_broadcast_pkts += REG_RD(bm_dev, mib.tx_broadcast_pkts);
	bm_dev->tx_multicast_pkts += REG_RD(bm_dev, mib.tx_multicast_pkts);
	bm_dev->tx_len_64 += REG_RD(bm_dev, mib.tx_len_64);
	bm_dev->tx_len_65_to_127 += REG_RD(bm_dev, mib.tx_len_65_to_127);
	bm_dev->tx_len_128_to_255 += REG_RD(bm_dev, mib.tx_len_128_to_255);
	bm_dev->tx_len_256_to_511 += REG_RD(bm_dev, mib.tx_len_256_to_511);
	bm_dev->tx_len_512_to_1023 += REG_RD(bm_dev, mib.tx_len_512_to_1023);
	bm_dev->tx_len_1024_to_max += REG_RD(bm_dev, mib.tx_len_1024_to_max);
	bm_dev->tx_jabber_pkts += REG_RD(bm_dev, mib.tx_jabber_pkts);
	bm_dev->tx_oversize_pkts += REG_RD(bm_dev, mib.tx_oversize_pkts);
	bm_dev->tx_fragment_pkts += REG_RD(bm_dev, mib.tx_fragment_pkts);
	bm_dev->tx_underruns += REG_RD(bm_dev, mib.tx_underruns);
	bm_dev->tx_total_cols += REG_RD(bm_dev, mib.tx_total_cols);
	bm_dev->tx_single_cols += REG_RD(bm_dev, mib.tx_single_cols);
	bm_dev->tx_multiple_cols += REG_RD(bm_dev, mib.tx_multiple_cols);
	bm_dev->tx_excessive_cols += REG_RD(bm_dev, mib.tx_excessive_cols);
	bm_dev->tx_late_cols += REG_RD(bm_dev, mib.tx_late_cols);
	bm_dev->tx_defered += REG_RD(bm_dev, mib.tx_defered);
	bm_dev->tx_carrier_lost += REG_RD(bm_dev, mib.tx_carrier_lost);
	bm_dev->tx_pause_pkts += REG_RD(bm_dev, mib.tx_pause_pkts);

	bm_dev->rx_good_octets += REG_RD(bm_dev, mib.rx_good_octets);
	bm_dev->rx_good_pkts += REG_RD(bm_dev, mib.rx_good_pkts);
	bm_dev->rx_octets += REG_RD(bm_dev, mib.rx_octets);
	bm_dev->rx_pkts += REG_RD(bm_dev, mib.rx_pkts);
	bm_dev->rx_broadcast_pkts += REG_RD(bm_dev, mib.rx_broadcast_pkts);
	bm_dev->rx_multicast_pkts += REG_RD(bm_dev, mib.rx_multicast_pkts);
	bm_dev->rx_len_64 += REG_RD(bm_dev, mib.rx_len_64);
	bm_dev->rx_len_65_to_127 += REG_RD(bm_dev, mib.rx_len_65_to_127);
	bm_dev->rx_len_128_to_255 += REG_RD(bm_dev, mib.rx_len_128_to_255);
	bm_dev->rx_len_256_to_511 += REG_RD(bm_dev, mib.rx_len_256_to_511);
	bm_dev->rx_len_512_to_1023 += REG_RD(bm_dev, mib.rx_len_512_to_1023);
	bm_dev->rx_len_1024_to_max += REG_RD(bm_dev, mib.rx_len_1024_to_max);
	bm_dev->rx_jabber_pkts += REG_RD(bm_dev, mib.rx_jabber_pkts);
	bm_dev->rx_oversize_pkts += REG_RD(bm_dev, mib.rx_oversize_pkts);
	bm_dev->rx_fragment_pkts += REG_RD(bm_dev, mib.rx_fragment_pkts);
	bm_dev->rx_missed_pkts += REG_RD(bm_dev, mib.rx_missed_pkts);
	bm_dev->rx_crc_align_errs += REG_RD(bm_dev, mib.rx_crc_align_errs);
	bm_dev->rx_undersize += REG_RD(bm_dev, mib.rx_undersize);
	bm_dev->rx_crc_errs += REG_RD(bm_dev, mib.rx_crc_errs);
	bm_dev->rx_align_errs += REG_RD(bm_dev, mib.rx_align_errs);
	bm_dev->rx_symbol_errs += REG_RD(bm_dev, mib.rx_symbol_errs);
	bm_dev->rx_pause_pkts += REG_RD(bm_dev, mib.rx_pause_pkts);
	bm_dev->rx_nonpause_pkts += REG_RD(bm_dev, mib.rx_nonpause_pkts);

	return 0;
}

void
b44_LM_WriteCam(struct bm4400_hw *bm_dev, u8 *ea, u32 camindex)
{
	u32 w;

	w = ((u32)ea[2] << 24) | ((u32)ea[3] << 16) |
		((u32) ea[4] << 8) | ea[5];
	REG_WR(bm_dev, camdatalo, w);
	w = CD_V | ((u32)ea[0] << 8) | ea[1];
	REG_WR(bm_dev, camdatahi, w);
	REG_WR(bm_dev, camcontrol, (((u32) camindex << CC_INDEX_SHIFT) |
		CC_WR));

	/* spin until done */
	SPINWAIT((REG_RD(bm_dev, camcontrol) & CC_CB), 100);
}

/******************************************************************************/
/* Description:                                                               */
/*                                                                            */
/* Return:                                                                    */
/******************************************************************************/
static u32
b44_GetPhyAdFlowCntrlSettings(
    struct bm4400_hw * bm_dev)
{
    u32 val;

    val = 0;

    /* Auto negotiation flow control only when autonegotiation is enabled. */
    if(bm_dev->DisableAutoNeg == FALSE ||
        bm_dev->RequestedLineSpeed == LM_LINE_SPEED_AUTO)
    {
        /* Please refer to Table 28B-3 of the 802.3ab-1999 spec. */
        if((bm_dev->FlowControlCap == LM_FLOW_CONTROL_AUTO_PAUSE) ||
            ((bm_dev->FlowControlCap & LM_FLOW_CONTROL_RECEIVE_PAUSE) &&
            (bm_dev->FlowControlCap & LM_FLOW_CONTROL_TRANSMIT_PAUSE)))
        {
            val |= PHY_AN_AD_PAUSE_CAPABLE;
        }
        else if(bm_dev->FlowControlCap & LM_FLOW_CONTROL_TRANSMIT_PAUSE)
        {
            val |= PHY_AN_AD_ASYM_PAUSE;
        }
        else if(bm_dev->FlowControlCap & LM_FLOW_CONTROL_RECEIVE_PAUSE)
        {
            val |= PHY_AN_AD_PAUSE_CAPABLE | PHY_AN_AD_ASYM_PAUSE;
        }
    }

    return val;
}


int
b44_LM_GetStats(struct bm4400_hw * bm_dev)
{
    return 0;
}

void
b44_LM_PollLink(struct bm4400_hw *bm_dev)
{
	u32 status, aux;
	u32 txcontrol;
	u32 LocalAdv, RemoteAdv;

	b44_LM_ReadPhy(bm_dev, 1, &status);
	b44_LM_ReadPhy(bm_dev, 1, &status);

	b44_LM_ReadPhy(bm_dev, 24, &aux);

	/* check for bad mdio read */
	if (status == 0xffff) {
		return;
	}

	/* update current speed and duplex */
	if (aux & AUX_SPEED)
		bm_dev->LineSpeed = LM_LINE_SPEED_100MBPS;
	else
		bm_dev->LineSpeed = LM_LINE_SPEED_10MBPS;
	if (aux & AUX_DUPLEX)
		bm_dev->DuplexMode = LM_DUPLEX_MODE_FULL;
	else
		bm_dev->DuplexMode = LM_DUPLEX_MODE_HALF;

	/* monitor link state */
	if ((bm_dev->LinkStatus == LM_STATUS_LINK_DOWN) &&
		(status & STAT_LINK)) {

		/* keep emac txcontrol duplex bit consistent with current */
		/* phy duplex */
		txcontrol = REG_RD(bm_dev, txcontrol);
		if ((bm_dev->DuplexMode == LM_DUPLEX_MODE_FULL) &&
			!(txcontrol & EXC_FD)) {

			REG_OR(bm_dev, txcontrol, EXC_FD);
		}
		else if ((bm_dev->DuplexMode == LM_DUPLEX_MODE_HALF) &&
			(txcontrol & EXC_FD)) {

			REG_AND(bm_dev, txcontrol, ~EXC_FD);
		}
		if (!bm_dev->DisableAutoNeg && (bm_dev->DuplexMode ==
			LM_DUPLEX_MODE_FULL)) {

			b44_LM_ReadPhy(bm_dev, PHY_AN_AD_REG, &LocalAdv);
			b44_LM_ReadPhy(bm_dev, PHY_LINK_PARTNER_ABILITY_REG,
				&RemoteAdv);
			b44_LM_SetFlowControl(bm_dev, LocalAdv, RemoteAdv);
		}

		bm_dev->LinkStatus = LM_STATUS_LINK_ACTIVE;
		b44_MM_IndicateStatus(bm_dev, LM_STATUS_LINK_ACTIVE);
	}
	else if ((bm_dev->LinkStatus == LM_STATUS_LINK_ACTIVE) &&
		!(status & STAT_LINK)) {

		bm_dev->LinkStatus = LM_STATUS_LINK_DOWN;
		b44_MM_IndicateStatus(bm_dev, LM_STATUS_LINK_DOWN);
	}


	/* check for remote fault error */
	if (status & STAT_REMFAULT) {
		printf("remote fault\n");
	}

	/* check for jabber error */
	if (status & STAT_JAB) {
		printf("jabber\n");
	}
}

/* reset and re-enable a core */
void
b44_LM_sb_core_reset(struct bm4400_hw *bm_dev)
{
	volatile u32 dummy;

	/*
	 * Must do the disable sequence first to work for arbitrary current core state.
	 */
	b44_LM_sb_core_disable(bm_dev);

	/*
	 * Now do the initialization sequence.
	 */

	/* set reset while enabling the clock and forcing them on throughout the core */
	REG_WR(bm_dev, sbconfig.sbtmstatelow,
		(SBTML_FGC | SBTML_CLK | SBTML_RESET));

	dummy = REG_RD(bm_dev, sbconfig.sbtmstatelow);
	b44_MM_Wait(1);

	/* PR3158 workaround - not fixed in any chip yet */
	if (REG_RD(bm_dev, sbconfig.sbtmstatehigh) & SBTMH_SERR) {
		printf("SBTMH_SERR; clearing...\n");
		REG_WR(bm_dev, sbconfig.sbtmstatehigh, 0);
		ASSERT(0);
	}
	if ((dummy = REG_RD(bm_dev, sbconfig.sbimstate)) &
		(SBIM_IBE | SBIM_TO)) {

		REG_AND(bm_dev, sbconfig.sbimstate, ~(SBIM_IBE | SBIM_TO));
		ASSERT(0);
	}

	/* clear reset and allow it to propagate throughout the core */
	REG_WR(bm_dev, sbconfig.sbtmstatelow, (SBTML_FGC | SBTML_CLK));
	dummy = REG_RD(bm_dev, sbconfig.sbtmstatelow);
	b44_MM_Wait(1);

	/* leave clock enabled */
	REG_WR(bm_dev, sbconfig.sbtmstatelow, SBTML_CLK);
	dummy = REG_RD(bm_dev, sbconfig.sbtmstatelow);
	b44_MM_Wait(1);
}

void
b44_LM_sb_core_disable(struct bm4400_hw *bm_dev)
{
	volatile u32 dummy;

	/* must return if core is already in reset */
	if (REG_RD(bm_dev, sbconfig.sbtmstatelow) & SBTML_RESET)
		return;
	
	/* set the reject bit */
	REG_WR(bm_dev, sbconfig.sbtmstatelow, (SBTML_CLK | SBTML_REJ));

	/* spin until reject is set */
	while ((REG_RD(bm_dev, sbconfig.sbtmstatelow) & SBTML_REJ) == 0)
		b44_MM_Wait(1);

	/* spin until sbtmstatehigh.busy is clear */
	while (REG_RD(bm_dev, sbconfig.sbtmstatehigh) & SBTMH_BUSY)
		b44_MM_Wait(1);

	/* set reset and reject while enabling the clocks */
	REG_WR(bm_dev, sbconfig.sbtmstatelow,
		(SBTML_FGC | SBTML_CLK | SBTML_REJ | SBTML_RESET));

	dummy = REG_RD(bm_dev, sbconfig.sbtmstatelow);
	b44_MM_Wait(10);

	/* leave reset and reject asserted */
	REG_WR(bm_dev, sbconfig.sbtmstatelow, (SBTML_REJ | SBTML_RESET));
	b44_MM_Wait(1);
}

/*
 * Configure the pci core for pci client (NIC) action
 * and return the pci core revision.
 */
u32
b44_LM_sb_pci_setup(struct bm4400_hw *bm_dev, u32 cores)
{
	u32 bar0window;
	sbpciregs_t *pciregs;
	u32 pcirev;

	pciregs = (sbpciregs_t *) bm_dev->mmio;

	/* save bar0window */
	b44_MM_ReadConfig32(bm_dev, PCI_BAR0_WIN, &bar0window);
	/* point bar0 at pci core registers */
	b44_MM_WriteConfig32(bm_dev, PCI_BAR0_WIN, b44_LM_getsbaddr(bm_dev,
		SBID_REG_PCI, 0));

	ASSERT(b44_LM_sb_coreid(bm_dev) == SB_PCI);

	pcirev = b44_LM_sb_corerev(bm_dev);

	/* enable sb->pci interrupts */
	REG_OR(bm_dev, sbconfig.sbintvec, cores);

	/* enable prefetch and bursts for sonics-to-pci translation 2 */
	REG_WR_OFFSET(bm_dev, OFFSETOF(sbpciregs_t, sbtopci2),
		REG_RD_OFFSET(bm_dev, OFFSETOF(sbpciregs_t, sbtopci2)) |
		(SBTOPCI_PREF|SBTOPCI_BURST));

	/* restore bar0window */
	b44_MM_WriteConfig32(bm_dev, PCI_BAR0_WIN, bar0window);

	return (pcirev);
}

/*
 * Return the SB address corresponding to core <id> instance <coreunit>.
 * Provide a layer of indirection between SB address map elements
 * and the individual chip maps.
 */
u32
b44_LM_getsbaddr(struct bm4400_hw *bm_dev, u32 id, u32 coreunit)
{
	struct sbmap *sbmap;
	int i;

	sbmap = bm_dev->sbmap;
	ASSERT(sbmap);

	for (i = 0; i < SBID_MAX; i++)
		if ((id == sbmap[i].id) && (coreunit == sbmap[i].coreunit))
			return (sbmap[i].sbaddr);

	ASSERT(0);
	return (0xdeadbeef);
}

u32
b44_LM_sb_base(u32 admatch)
{
	u32 base;
	u32 type;

	type = admatch & SBAM_TYPE_MASK;
	ASSERT(type < 3);

	base = 0;

	if (type == 0) {
		base = admatch & SBAM_BASE0_MASK;
	} else if (type == 1) {
		ASSERT(admatch & SBAM_ADEN);
		ASSERT(!(admatch & SBAM_ADNEG));	/* neg not supported */
		base = admatch & SBAM_BASE1_MASK;
	} else if (type == 2) {
		ASSERT(admatch & SBAM_ADEN);
		ASSERT(!(admatch & SBAM_ADNEG));	/* neg not supported */
		base = admatch & SBAM_BASE2_MASK;
	}

	return (base);
}

u32
b44_LM_sb_size(u32 admatch)
{
	u32 size;
	u32 type;

	type = admatch & SBAM_TYPE_MASK;
	ASSERT(type < 3);

	size = 0;

	if (type == 0) {
		size = 1 << (((admatch & SBAM_ADINT0_MASK) >> SBAM_ADINT0_SHIFT) + 1);
	} else if (type == 1) {
		ASSERT(admatch & SBAM_ADEN);
		ASSERT(!(admatch & SBAM_ADNEG));	/* neg not supported */
		size = 1 << (((admatch & SBAM_ADINT1_MASK) >> SBAM_ADINT1_SHIFT) + 1);
	} else if (type == 2) {
		ASSERT(admatch & SBAM_ADEN);
		ASSERT(!(admatch & SBAM_ADNEG));	/* neg not supported */
		size = 1 << (((admatch & SBAM_ADINT2_MASK) >> SBAM_ADINT2_SHIFT) + 1);
	}

	return (size);
}

/* return the core instance number of this core */
u32
b44_LM_sb_coreunit(struct bm4400_hw *bm_dev)
{
	struct sbmap *sbmap;
	u32 base;
	int i;

	sbmap = bm_dev->sbmap;
	ASSERT(sbmap);

	base = b44_LM_sb_base(REG_RD(bm_dev, sbconfig.sbadmatch0));

	for (i = 0; i < SBID_MAX; i++)
		if (base == sbmap[i].sbaddr)
			return (sbmap[i].coreunit);

	ASSERT(0);
	return (0xdeadbeef);
}


u32
b44_LM_sb_clock(struct bm4400_hw *bm_dev, u32 extifva)
{
	ASSERT(0);	/* XXX TBD */
	return (0);
}

u32
b44_LM_sb_coreid(struct bm4400_hw *bm_dev)
{
	return ((REG_RD(bm_dev, sbconfig.sbidhigh) &
		SBIDH_CC_MASK) >> SBIDH_CC_SHIFT);
}

u32
b44_LM_sb_corerev(struct bm4400_hw *bm_dev)
{
	return (REG_RD(bm_dev, sbconfig.sbidhigh) & SBIDH_RC_MASK);
}

u32
b44_LM_sb_iscoreup(struct bm4400_hw *bm_dev)
{
	return ((REG_RD(bm_dev, sbconfig.sbtmstatelow) &
		(SBTML_RESET | SBTML_REJ | SBTML_CLK)) == SBTML_CLK);
}

#ifdef BCM_WOL

/* Program patterns on the chip */
static void
b44_LM_pmprog(struct bm4400_hw *bm_dev)
{
	u32 wfl;
	int plen0, plen1, max, i, j;
	u8 wol_pattern[BCMENET_PMPSIZE];
	u8 wol_mask[BCMENET_PMMSIZE];

	/* program the chip with wakeup patterns, masks, and lengths */

	if (bm_dev->WakeUpMode == LM_WAKE_UP_MODE_NONE) {
   		wfl = DISABLE_3210_PATMATCH;  
		REG_WR(bm_dev, wakeuplength, wfl);
	}
	else if (bm_dev->WakeUpMode == LM_WAKE_UP_MODE_MAGIC_PACKET) {
		/* UDP magic packet pattern */
		memset(wol_pattern, 0, BCMENET_PMPSIZE);
		memset(wol_pattern + 42, 0xff, 6);	/* sync pattern */
		max = ETHERNET_ADDRESS_SIZE;
		for (i = 0; i < 14; ++i) {
			if (i == 13)
				max = 2;
			for (j = 0; j < max; ++j) {
				wol_pattern[42 + 6 +
				(i * ETHERNET_ADDRESS_SIZE) + j] =
				bm_dev->NodeAddress[j];
			}
		}
		memset(wol_mask, 0, BCMENET_PMMSIZE);
		wol_mask[5] = 0xfc;
		memset(wol_mask + 6, 0xff, 10);
		plen0 = BCMENET_PMPSIZE - 1;

   		b44_LM_ftwrite(bm_dev, (u32 *)wol_pattern,
			BCMENET_PMPSIZE, BCMENET_PMPBASE);

   		b44_LM_ftwrite(bm_dev, (u32 *)wol_mask, BCMENET_PMMSIZE,
    			BCMENET_PMMBASE);

		/* raw ethernet II magic packet pattern */
		memset(wol_pattern, 0, BCMENET_PMPSIZE);
		memset(wol_pattern + 14, 0xff, 6);	/* sync pattern */
		max = ETHERNET_ADDRESS_SIZE;
		for (i = 0; i < 16; ++i) {
			for (j = 0; j < max; ++j) {
				wol_pattern[14 + 6 +
				(i * ETHERNET_ADDRESS_SIZE) + j] =
				bm_dev->NodeAddress[j];
			}
		}
		memset(wol_mask, 0, BCMENET_PMMSIZE);
		wol_mask[2] = 0xf0;
		memset(wol_mask + 3, 0xff, 11);
		wol_mask[14] = 0xf;
		plen1 = 14 + 6 + 96 - 1;

   		b44_LM_ftwrite(bm_dev, (u32 *)wol_pattern,
			BCMENET_PMPSIZE, BCMENET_PMPBASE + BCMENET_PMPSIZE);

   		b44_LM_ftwrite(bm_dev, (u32 *)wol_mask, BCMENET_PMMSIZE,
    			BCMENET_PMMBASE + BCMENET_PMMSIZE);

   		/* set this pattern's length: one less than the real length */
   		wfl = plen0 | (plen1 << 8) | DISABLE_32_PATMATCH;

		REG_WR(bm_dev, wakeuplength, wfl);
	}

}

void
b44_LM_pmset(struct bm4400_hw *bm_dev)
{
	u16 Value16;

	b44_LM_Halt(bm_dev);

	/* now turn on just enough of the chip to receive and match patterns */
	b44_LM_ResetAdapter(bm_dev, FALSE);

	/* program patterns */
	b44_LM_pmprog(bm_dev);

	/* enable chip wakeup pattern matching */
	REG_OR(bm_dev, devcontrol, DC_PM);
	/* enable sonics bus PME */
	REG_OR(bm_dev, sbconfig.sbtmstatelow, SBTML_PE);

	b44_MM_ReadConfig16(bm_dev, BCMENET_PMCSR, &Value16);
	b44_MM_WriteConfig16(bm_dev, BCMENET_PMCSR,
		Value16 | ENABLE_PCICONFIG_PME);
}


static void
b44_LM_ftwrite(struct bm4400_hw *bm_dev, u32 *b, u32 nbytes,
	u32 ftaddr)
{
	u32 i;

	for (i = 0; i < nbytes; i += sizeof(u32)) {
		REG_WR(bm_dev, enetftaddr, ftaddr + i);
		REG_WR(bm_dev, enetftdata, b[i / sizeof(u32)]);
	}
}

#endif /* BCM_WOL */
