
/******************************************************************************/
/*                                                                            */
/* Broadcom BCM4400 Linux Network Driver, Copyright (c) 2002 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/******************************************************************************/

#ifndef B44MM_H
#define B44MM_H

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <asm/processor.h>		/* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/unaligned.h>
#include <linux/delay.h>
#include <asm/byteorder.h>
#include <linux/time.h>
#include <linux/ethtool.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>

#ifdef CONFIG_PROC_FS
#define BCM_PROC_FS 1
#endif

#define BCM_WOL 1

#ifdef __BIG_ENDIAN
#define BIG_ENDIAN_HOST 1
#endif

#define MM_SWAP_LE16(x) cpu_to_le16(x)

#define MM_MEMWRITEL(ptr, val) __raw_writel(val, ptr)
#define MM_MEMREADL(ptr) __raw_readl(ptr)

#define MM_ATOMIC_SET(ptr, val) atomic_set(ptr, val)
#define MM_ATOMIC_READ(ptr) atomic_read(ptr)
#define MM_ATOMIC_INC(ptr) atomic_inc(ptr)
#define MM_ATOMIC_ADD(ptr, val) atomic_add(val, ptr)
#define MM_ATOMIC_DEC(ptr) atomic_dec(ptr)
#define MM_ATOMIC_SUB(ptr, val) atomic_sub(val, ptr)

#define MM_MB() mb()
#define MM_WMB() wmb()

#include "b44lm.h"
#include "b44queue.h"
#include "b44.h"

#if DBG
#define static
#else
#define static static
#endif

extern int b44_Packet_Desc_Size;

#define B44_MM_PACKET_DESC_SIZE b44_Packet_Desc_Size

DECLARE_QUEUE_TYPE(UM_RX_PACKET_Q, MAX_RX_PACKET_DESC_COUNT+1);

#define MAX_MEM 16

#if MAX_SKB_FRAGS

typedef dma_addr_t dmaaddr_high_t;

#endif /* #if MAX_SKB_FRAGS*/

#if defined (CONFIG_X86)
#define NO_PCI_UNMAP 1
#endif

struct bm4400_dev {
	struct bm4400_hw lm_dev;
	struct net_device *dev;
	struct pci_dev *pdev;
	struct net_device *next_module;
	char *name;
#ifdef BCM_PROC_FS
	struct proc_dir_entry *pfs_entry;
	char pfs_name[32];
#endif
	void *mem_list[MAX_MEM];
	dma_addr_t dma_list[MAX_MEM];
	int mem_size_list[MAX_MEM];
	int mem_list_num;
	int index;
	int opened;
	int delayed_link_ind; /* Delay link status during initial load */
	int timer_interval;
	int link_interval;
	int tx_full;
	int tx_queued;
	int line_speed;		/* in Mbps, 0 if link is down */
	UM_RX_PACKET_Q rx_out_of_buf_q;
	int rx_buf_repl_thresh;
	int rx_buf_repl_panic_thresh;
	struct timer_list timer;
	spinlock_t phy_lock;
	volatile unsigned long interrupt;
	atomic_t intr_sem;
	int tasklet_pending;
	volatile unsigned long tasklet_busy;
	struct tasklet_struct tasklet;
	struct net_device_stats stats;
	uint rx_misc_errors;
};

struct bm4400_packet_buf {
	struct bm4400_packet lm_packet;
	struct sk_buff *skbuff;
#if MAX_SKB_FRAGS
	DECLARE_PCI_UNMAP_ADDR(map[MAX_SKB_FRAGS + 1])
	DECLARE_PCI_UNMAP_LEN(map_len[MAX_SKB_FRAGS + 1])
#else
	DECLARE_PCI_UNMAP_ADDR(map[1])
	DECLARE_PCI_UNMAP_LEN(map_len[1])
#endif
};


static inline void b44_MM_MapRxDma(struct bm4400_hw * bm_dev,
	struct bm4400_packet *packet,
	u32 *paddr)
{
	dma_addr_t map;
	struct bm4400_dev *dev = (struct bm4400_dev *)bm_dev;

	map = pci_map_single(dev->pdev,
			packet->u.Rx.pRxBufferVirt,
			packet->u.Rx.RxBufferSize,
			PCI_DMA_FROMDEVICE);
	pci_unmap_addr_set(packet, map[0], map);
	*paddr = (u32) map;
}

static inline void b44_MM_MapTxDma(struct bm4400_hw * bm_dev,
	struct bm4400_packet *packet,
	u32 *paddr, u32 *len, int frag)
{
	dma_addr_t map;
	struct bm4400_dev *dev = (struct bm4400_dev *)bm_dev;
	struct bm4400_packet_buf *buf = (struct bm4400_packet_buf *)packet;
	struct sk_buff *skb = buf->skbuff;

	map = pci_map_single(dev->pdev,
		skb->data,
		skb->len,
		PCI_DMA_TODEVICE);
	pci_unmap_addr_set(packet, map[0], map);
	*paddr = (u32) map;
	*len = skb->len;
}

#define B44_MM_PTR(_ptr)   ((unsigned long) (_ptr))

#if (BITS_PER_LONG == 64)
#define B44_MM_GETSTATS(_Ctr) \
	(unsigned long) (_Ctr).Low + ((unsigned long) (_Ctr).High << 32)
#else
#define B44_MM_GETSTATS(_Ctr) \
	(unsigned long) (_Ctr).Low
#endif


#define printf(fmt, args...) printk(KERN_DEBUG fmt, ##args)

#define DbgPrint(fmt, arg...) printk(KERN_WARNING fmt, ##arg)
#if defined(CONFIG_X86)
#define DbgBreakPoint() __asm__("int $129")
#else
#define DbgBreakPoint()
#endif
#define b44_MM_Wait(time) udelay(time)

#define ASSERT(expr)							\
	if (!(expr)) {							\
		printk(KERN_DEBUG "ASSERT failed: %s\n", #expr);	\
	}
#endif
