/******************************************************************************/
/*                                                                            */
/* Broadcom BCM4400 Linux Network Driver, Copyright (c) 2002 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/* Change Log                                                                 */
/* 1.0.3  (2/25/03)                                                           */
/*  - Fixed various problems related to reset.                                */
/*  - Added magic packet WOL.                                                 */
/******************************************************************************/


char bcm4400_driver[] = "bcm4400";
char bcm4400_version[] = "2.0.0";
char bcm4400_date[] = "(03/25/03)";

#define B44UM
#include "b44mm.h"

/* A few user-configurable values. */

#define MAX_UNITS 16
/* Used to pass the full-duplex flag, etc. */
static int line_speed[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int auto_speed[MAX_UNITS] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
static int full_duplex[MAX_UNITS] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
static int rx_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int tx_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int auto_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

#define TX_DESC_CNT DEFAULT_TX_PACKET_DESC_COUNT
static unsigned int tx_pkt_desc_cnt[MAX_UNITS] =
	{TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,
	TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,
	TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,TX_DESC_CNT,
	TX_DESC_CNT};

#define RX_DESC_CNT DEFAULT_RX_PACKET_DESC_COUNT
static unsigned int rx_pkt_desc_cnt[MAX_UNITS] =
	{RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,
	RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,
	RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,RX_DESC_CNT,
	RX_DESC_CNT };

#ifdef BCM_WOL
static int enable_wol[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#endif

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (2*HZ)

#define BCM4400_PHY_LOCK(pUmDevice, flags)			\
{								\
	spin_lock_irqsave(&(pUmDevice)->phy_lock, flags);	\
}

#define BCM4400_PHY_UNLOCK(pUmDevice, flags)			\
{								\
	spin_unlock_irqrestore(&(pUmDevice)->phy_lock, flags);	\
}

void
bcm4400_intr_off(struct bm4400_dev * pUmDevice)
{
	atomic_inc(&pUmDevice->intr_sem);
	b44_LM_DisableInterrupt(&pUmDevice->lm_dev);
	synchronize_irq();
}

void
bcm4400_intr_on(struct bm4400_dev * pUmDevice)
{
	if (atomic_dec_and_test(&pUmDevice->intr_sem)) {
		b44_LM_EnableInterrupt(&pUmDevice->lm_dev);
	}
}

int b44_Packet_Desc_Size = sizeof(struct bm4400_packet_buf);

MODULE_AUTHOR("Michael Chan <mchan@broadcom.com>");
MODULE_DESCRIPTION("BCM4400 Driver");
MODULE_LICENSE("GPL");
MODULE_PARM(line_speed, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(auto_speed, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(rx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(tx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(auto_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(tx_pkt_desc_cnt, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(rx_pkt_desc_cnt, "1-" __MODULE_STRING(MAX_UNITS) "i");
#ifdef BCM_WOL
MODULE_PARM(enable_wol, "1-" __MODULE_STRING(MAX_UNITS) "i");
#endif

static int bcm4400_open(struct net_device *dev);
static void bcm4400_timer(unsigned long data);
static void bcm4400_tx_timeout(struct net_device *dev);
static int bcm4400_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void bcm4400_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
#ifdef BCM_TASKLET
static void bcm4400_tasklet(unsigned long data);
#endif
static int bcm4400_close(struct net_device *dev);
static struct net_device_stats *bcm4400_get_stats(struct net_device *dev);
static int bcm4400_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static void bcm4400_set_rx_mode(struct net_device *dev);
static int bcm4400_set_mac_addr(struct net_device *dev, void *p);
static int bcm4400_rxfill(struct bm4400_dev * pUmDevice);
static int bcm4400_freemem(struct net_device *dev);
static void bcm4400_free_remaining_rx_bufs(struct bm4400_dev *pUmDevice);


/* A list of all installed bcm4400 devices. */
static struct net_device *root_bcm4400_dev = NULL;

typedef enum {
	BCM4401 = 0,
} board_t;


/* indexed by board_t, above */
static struct {
	char *name;
} board_info[] __devinitdata = {
	{ "Broadcom BCM4401 100Base-T" },
	{ 0 },
	};

static struct pci_device_id bcm4400_pci_tbl[] __devinitdata = {
	{0x14e4, 0x4401, PCI_ANY_ID, PCI_ANY_ID, 0, 0, BCM4401 },
	{0,}
};

MODULE_DEVICE_TABLE(pci, bcm4400_pci_tbl);

#ifdef BCM_PROC_FS
extern int bcm4400_proc_create(void);
extern int bcm4400_proc_create_dev(struct net_device *dev);
extern int bcm4400_proc_remove_dev(struct net_device *dev);
#endif

static int __devinit bcm4400_init_board(struct pci_dev *pdev,
					struct net_device **dev_out,
					int board_idx)
{
	struct net_device *dev;
	struct bm4400_dev * pUmDevice;
	struct bm4400_hw * bm_dev;
	int rc;

	*dev_out = NULL;

	/* dev zeroed in init_etherdev */
	dev = init_etherdev(NULL, sizeof(*pUmDevice));
	if (dev == NULL) {
		printk (KERN_ERR "%s: unable to alloc new ethernet\n",
			bcm4400_driver);
		return -ENOMEM;
	}
	SET_MODULE_OWNER(dev);
	pUmDevice = (struct bm4400_dev *) dev->priv;

	/* enable device (incl. PCI PM wakeup), and bus-mastering */
	rc = pci_enable_device (pdev);
	if (rc)
		goto err_out;

	rc = pci_request_regions(pdev, bcm4400_driver);
	if (rc)
		goto err_out;

	pci_set_master(pdev);

	if (pci_set_dma_mask(pdev, ~(0UL)) != 0) {
		printk(KERN_ERR "System does not support DMA\n");
		pci_release_regions(pdev);
		goto err_out;
	}

	spin_lock_init(&pUmDevice->phy_lock);

	pUmDevice->dev = dev;
	pUmDevice->pdev = pdev;
	pUmDevice->mem_list_num = 0;
	pUmDevice->next_module = root_bcm4400_dev;
	pUmDevice->index = board_idx;
	root_bcm4400_dev = dev;

	bm_dev = (struct bm4400_hw *) pUmDevice;

	if (b44_LM_GetAdapterInfo(bm_dev) != 0) {
		printk(KERN_ERR "Get Adapter info failed\n");
		rc = -ENODEV;
		goto err_out_unmap;
	}

	dev->mem_start = pci_resource_start(pdev, 0);
	dev->mem_end = dev->mem_start + sizeof(bcmenetregs_t) + 128; 
	dev->irq = pdev->irq;

	*dev_out = dev;
	return 0;

err_out_unmap:
	pci_release_regions(pdev);
	bcm4400_freemem(dev);

err_out:
	unregister_netdev(dev);
	kfree (dev);
	return rc;
}

static int __devinit
bcm4400_print_ver(void)
{
	printk(KERN_INFO "Broadcom 4401 Ethernet Driver %s ",
		bcm4400_driver);
	printk("ver. %s %s\n", bcm4400_version, bcm4400_date);
	return 0;
}

static int __devinit
bcm4400_init_one(struct pci_dev *pdev,
				       const struct pci_device_id *ent)
{
	struct net_device *dev = NULL;
	struct bm4400_dev * pUmDevice;
	struct bm4400_hw * bm_dev;
	int i;
	static int board_idx = -1;
	static int printed_version = 0;
	struct pci_dev *amd_dev;

	board_idx++;

	if (!printed_version) {
		bcm4400_print_ver();
#ifdef BCM_PROC_FS
		bcm4400_proc_create();
#endif
		printed_version = 1;
	}

	i = bcm4400_init_board(pdev, &dev, board_idx);
	if (i < 0) {
		return i;
	}

	if (dev == NULL)
		return -ENOMEM;

	dev->open = bcm4400_open;
	dev->hard_start_xmit = bcm4400_start_xmit;
	dev->stop = bcm4400_close;
	dev->get_stats = bcm4400_get_stats;
	dev->set_multicast_list = bcm4400_set_rx_mode;
	dev->do_ioctl = bcm4400_ioctl;
	dev->set_mac_address = &bcm4400_set_mac_addr;
	dev->tx_timeout = bcm4400_tx_timeout;
	dev->watchdog_timeo = TX_TIMEOUT;

	pUmDevice = (struct bm4400_dev *) dev->priv;
	bm_dev = (struct bm4400_hw *) pUmDevice;

	dev->base_addr = pci_resource_start(pdev, 0);
	dev->irq = pdev->irq;

	pci_set_drvdata(pdev, dev);

	memcpy(dev->dev_addr, bm_dev->NodeAddress, 6);
	pUmDevice->name = board_info[ent->driver_data].name,
	printk(KERN_INFO "%s: %s found at mem %lx, IRQ %d, ",
		dev->name, pUmDevice->name, dev->base_addr,
		dev->irq);
	printk("node addr ");
	for (i = 0; i < 6; i++) {
		printk("%2.2x", dev->dev_addr[i]);
	}
	printk("\n");

#ifdef BCM_PROC_FS
	bcm4400_proc_create_dev(dev);
#endif
#ifdef BCM_TASKLET
	tasklet_init(&pUmDevice->tasklet, bcm4400_tasklet,
		(unsigned long) pUmDevice);
#endif
	if ((amd_dev = pci_find_device(0x1022, 0x700c, NULL))) {
		u32 val;

		/* Found AMD 762 North bridge */
		pci_read_config_dword(amd_dev, 0x4c, &val);
		if ((val & 0x02) == 0) {
			pci_write_config_dword(amd_dev, 0x4c, val | 0x02);
			printk(KERN_INFO "%s: Setting AMD762 Northbridge to enable PCI ordering compliance\n", bcm4400_driver);
		}
	}
	return 0;

}


static void __devexit
bcm4400_remove_one (struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata (pdev);
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;

#ifdef BCM_PROC_FS
	bcm4400_proc_remove_dev(dev); 
#endif
	unregister_netdev(dev);

	if (pUmDevice->lm_dev.pMappedMemBase)
		iounmap(pUmDevice->lm_dev.pMappedMemBase);

	pci_release_regions(pdev);

	kfree(dev);

	pci_set_drvdata(pdev, NULL);

/*	pci_power_off(pdev, -1);*/

}



static int
bcm4400_open(struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;

	pUmDevice->delayed_link_ind = (4 * HZ) / pUmDevice->timer_interval;

	if (request_irq(dev->irq, &bcm4400_interrupt, SA_SHIRQ, dev->name, dev)) {
		return -EAGAIN;
	}

	pUmDevice->opened = 1;
	if (b44_LM_InitializeAdapter(bm_dev) != 0) {
		free_irq(dev->irq, dev);
		bcm4400_freemem(dev);
		return -EAGAIN;
	}

	if (memcmp(dev->dev_addr, bm_dev->NodeAddress, 6)) {
		b44_LM_SetMacAddress(bm_dev, dev->dev_addr);
	}

	QQ_InitQueue(&pUmDevice->rx_out_of_buf_q.Container,
        MAX_RX_PACKET_DESC_COUNT);
	netif_start_queue(dev);

	init_timer(&pUmDevice->timer);
	pUmDevice->timer.expires = jiffies + pUmDevice->timer_interval;
	pUmDevice->timer.data = (unsigned long)dev;
	pUmDevice->timer.function = &bcm4400_timer;
	add_timer(&pUmDevice->timer);

	atomic_set(&pUmDevice->intr_sem, 0);
	b44_LM_EnableInterrupt(bm_dev);

	return 0;
}



static void
bcm4400_timer(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = &pUmDevice->lm_dev;

	if (!pUmDevice->opened)
		return;

	if (atomic_read(&pUmDevice->intr_sem)) {
		pUmDevice->timer.expires = jiffies + pUmDevice->timer_interval;
		add_timer(&pUmDevice->timer);
		return;
	}

	if (pUmDevice->delayed_link_ind > 0) {
		pUmDevice->delayed_link_ind--;
		if (pUmDevice->delayed_link_ind == 0) {
			b44_MM_IndicateStatus(bm_dev, bm_dev->LinkStatus);
		}
	}
	if (!pUmDevice->interrupt) {
		if (REG_RD(bm_dev, intstatus) & I_XI) {
			REG_WR(bm_dev, gptimer, 2);
		}
	}
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container) >
		pUmDevice->rx_buf_repl_panic_thresh) {
		/* Generate interrupt and let isr allocate buffers */
		REG_WR(bm_dev, gptimer, 2);
	}

	if (pUmDevice->link_interval == 0) {
		b44_LM_PollLink(bm_dev);
		if (bm_dev->LinkStatus == LM_STATUS_LINK_ACTIVE)
			b44_LM_StatsUpdate(bm_dev);
		pUmDevice->link_interval = HZ / pUmDevice->timer_interval;
	}
	else {
		pUmDevice->link_interval--;
	}

	b44_LM_GetStats(bm_dev);

	pUmDevice->timer.expires = jiffies + pUmDevice->timer_interval;
	add_timer(&pUmDevice->timer);
}

static void
bcm4400_tx_timeout(struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;

	netif_stop_queue(dev);
	bcm4400_intr_off(pUmDevice);
	b44_LM_ResetAdapter(bm_dev, TRUE);
	if (memcmp(dev->dev_addr, bm_dev->NodeAddress, 6)) {
		b44_LM_SetMacAddress(bm_dev, dev->dev_addr);
	}
	atomic_set(&pUmDevice->intr_sem, 1);
	bcm4400_intr_on(pUmDevice);
	netif_wake_queue(dev);
}

static int
bcm4400_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
	struct bm4400_packet * packet;
	struct bm4400_packet_buf * buf;

	if ((bm_dev->LinkStatus == LM_STATUS_LINK_DOWN) || !bm_dev->InitDone)
	{
		dev_kfree_skb(skb);
		return 0;
	}
	
	packet = (struct bm4400_packet *)
		QQ_PopHead(&bm_dev->TxPacketFreeQ.Container);
	if (packet == 0) {
		netif_stop_queue(dev);
		pUmDevice->tx_full = 1;
		if (QQ_GetEntryCnt(&bm_dev->TxPacketFreeQ.Container)) {
			netif_wake_queue(dev);
			pUmDevice->tx_full = 0;
		}
		return 1;
	}
	buf = (struct bm4400_packet_buf *) packet;
	buf->skbuff = skb;

	if (atomic_read(&bm_dev->SendDescLeft) == 0) {
		netif_stop_queue(dev);
		pUmDevice->tx_full = 1;
		QQ_PushHead(&bm_dev->TxPacketFreeQ.Container, packet);
		if (atomic_read(&bm_dev->SendDescLeft)) {
			netif_wake_queue(dev);
			pUmDevice->tx_full = 0;
		}
		return 1;
	}

	packet->u.Tx.FragCount = 1;

	b44_LM_SendPacket(bm_dev, packet);

	dev->trans_start = jiffies;
	return 0;
}


static void
bcm4400_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *)dev_instance;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
#ifdef BCM_TASKLET
	int repl_buf_count;
#endif

	if (!bm_dev->InitDone)
		return;

	if (atomic_read(&pUmDevice->intr_sem))
		return;

	if (test_and_set_bit(0, (void*)&pUmDevice->interrupt)) {
		printk(KERN_ERR "%s: Duplicate entry of the interrupt handler by "
			   "processor %d.\n",
			   dev->name, hard_smp_processor_id());
		return;
	}


	b44_LM_ServiceInterrupts(bm_dev);

#ifdef BCM_TASKLET
	repl_buf_count = QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container);
	if (repl_buf_count > pUmDevice->rx_buf_repl_thresh) {
		if ((repl_buf_count > pUmDevice->rx_buf_repl_panic_thresh) &&
			(!test_and_set_bit(0, &pUmDevice->tasklet_busy))) {
			bcm4400_rxfill(pUmDevice);
			clear_bit(0, (void*)&pUmDevice->tasklet_busy);
		}
		else if (!pUmDevice->tasklet_pending) {
			pUmDevice->tasklet_pending = 1;
			tasklet_schedule(&pUmDevice->tasklet);
		}
	}
#else
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container)) {
		bcm4400_rxfill(pUmDevice);
	}

	if (QQ_GetEntryCnt(&bm_dev->RxPacketFreeQ.Container)) {
		b44_LM_QueueRxPackets(bm_dev);
	}
#endif

	clear_bit(0, (void*)&pUmDevice->interrupt);
	if (pUmDevice->tx_queued) {
		pUmDevice->tx_queued = 0;
		netif_wake_queue(dev);
	}
	return;
}


#ifdef BCM_TASKLET
static void
bcm4400_tasklet(unsigned long data)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)data;

	/* RH 7.2 Beta 3 tasklets are reentrant */
	if (test_and_set_bit(0, &pUmDevice->tasklet_busy)) {
		pUmDevice->tasklet_pending = 0;
		return;
	}

	pUmDevice->tasklet_pending = 0;
	bcm4400_rxfill(pUmDevice);
	clear_bit(0, &pUmDevice->tasklet_busy);
}
#endif

static int
bcm4400_close(struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;

	netif_stop_queue(dev);
	pUmDevice->opened = 0;

	bcm4400_intr_off(pUmDevice);
	netif_carrier_off(dev);
#ifdef BCM_TASKLET
//	tasklet_disable(&pUmDevice->tasklet);
	tasklet_kill(&pUmDevice->tasklet);
#endif

	b44_LM_Halt(bm_dev);
	bm_dev->InitDone = 0;
	bcm4400_free_remaining_rx_bufs(pUmDevice);
	del_timer(&pUmDevice->timer);

	free_irq(dev->irq, dev);
#ifdef BCM_WOL
	if (bm_dev->WakeUpMode != LM_WAKE_UP_MODE_NONE) {
		b44_LM_pmset(bm_dev);
	}
#endif

	bcm4400_freemem(dev);

	return 0;
}

static void
bcm4400_free_remaining_rx_bufs(struct bm4400_dev *pUmDevice)
{
	struct bm4400_hw *bm_dev = &pUmDevice->lm_dev;
	struct bm4400_packet_buf *buf;
	struct sk_buff *skb;
	int cnt, i;

	cnt = QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container);
	for (i = 0; i < cnt; i++) {
		if ((buf =
			QQ_PopHead(&pUmDevice->rx_out_of_buf_q.Container))
			!= 0) {

			if ((skb = buf->skbuff))
				dev_kfree_skb(skb);
			buf->skbuff = 0;
			QQ_PushTail(&bm_dev->RxPacketFreeQ.Container,
				buf);
		}
	}
}

static int
bcm4400_freemem(struct net_device *dev)
{
	int i;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;

	for (i = 0; i < pUmDevice->mem_list_num; i++) {
		if (pUmDevice->mem_size_list[i] == 0) {
			kfree(pUmDevice->mem_list[i]);
		}
		else {
			pci_free_consistent(pUmDevice->pdev,
				(size_t) pUmDevice->mem_size_list[i],
				pUmDevice->mem_list[i],
				pUmDevice->dma_list[i]);
		}
	}
	pUmDevice->mem_list_num = 0;
	return 0;
}

static struct net_device_stats *
bcm4400_get_stats(struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw *bm_dev = &pUmDevice->lm_dev;
	struct net_device_stats *p_netstats = &pUmDevice->stats;

	p_netstats->rx_packets = bm_dev->rx_pkts;
	p_netstats->tx_packets = bm_dev->tx_pkts;
	p_netstats->rx_bytes = bm_dev->rx_octets;
	p_netstats->tx_bytes = bm_dev->tx_octets;
	p_netstats->tx_errors = bm_dev->tx_jabber_pkts +
		bm_dev->tx_oversize_pkts + bm_dev->tx_underruns +
		bm_dev->tx_excessive_cols + bm_dev->tx_late_cols;
	p_netstats->multicast = bm_dev->tx_multicast_pkts;
	p_netstats->collisions = bm_dev->tx_total_cols;
	p_netstats->rx_length_errors = bm_dev->rx_oversize_pkts +
		bm_dev->rx_undersize;
	p_netstats->rx_over_errors = bm_dev->rx_missed_pkts;
	p_netstats->rx_frame_errors = bm_dev->rx_align_errs;
	p_netstats->rx_crc_errors = bm_dev->rx_crc_errs;
	p_netstats->rx_errors = bm_dev->rx_jabber_pkts +
		bm_dev->rx_oversize_pkts + bm_dev->rx_missed_pkts +
		bm_dev->rx_crc_align_errs + bm_dev->rx_undersize +
		bm_dev->rx_crc_errs + bm_dev->rx_align_errs +
		bm_dev->rx_symbol_errs;
	p_netstats->tx_aborted_errors = bm_dev->tx_underruns;
	p_netstats->tx_carrier_errors = bm_dev->tx_carrier_lost;
	return p_netstats;
}

#ifdef SIOCETHTOOL
static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr)
{
	struct ethtool_cmd ethcmd;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
	unsigned long flags;
		
	if (copy_from_user(&ethcmd, useraddr, sizeof(ethcmd)))
		return -EFAULT;

        switch (ethcmd.cmd) {
#ifdef ETHTOOL_GDRVINFO
        case ETHTOOL_GDRVINFO: {
		struct ethtool_drvinfo info = {ETHTOOL_GDRVINFO};

		strcpy(info.driver,  bcm4400_driver);
		strcpy(info.version, bcm4400_version);
		strcpy(info.bus_info, pUmDevice->pdev->slot_name);
		if (copy_to_user(useraddr, &info, sizeof(info)))
			return -EFAULT;
		return 0;
	}
#endif
        case ETHTOOL_GSET: {
		ethcmd.supported =
			(SUPPORTED_10baseT_Half |
			SUPPORTED_10baseT_Full |
			SUPPORTED_100baseT_Half |
			SUPPORTED_100baseT_Full |
			SUPPORTED_Autoneg);
			ethcmd.supported |= SUPPORTED_TP;
			ethcmd.port = PORT_TP;

		ethcmd.transceiver = XCVR_INTERNAL;
		ethcmd.phy_address = 0;

		if (pUmDevice->line_speed == 100)
			ethcmd.speed = SPEED_100;
		else if (pUmDevice->line_speed == 10)
			ethcmd.speed = SPEED_10;
		else
			ethcmd.speed = 0;

		if (bm_dev->DuplexMode == LM_DUPLEX_MODE_FULL)
			ethcmd.duplex = DUPLEX_FULL;
		else
			ethcmd.duplex = DUPLEX_HALF;

		if (bm_dev->DisableAutoNeg == FALSE) {
			ethcmd.autoneg = AUTONEG_ENABLE;
			ethcmd.advertising = ADVERTISED_Autoneg;
			ethcmd.advertising |=
				ADVERTISED_TP;
			if (bm_dev->Advertising &
				PHY_AN_AD_10BASET_HALF) {

				ethcmd.advertising |=
					ADVERTISED_10baseT_Half;
			}
			if (bm_dev->Advertising &
				PHY_AN_AD_10BASET_FULL) {

				ethcmd.advertising |=
					ADVERTISED_10baseT_Full;
			}
			if (bm_dev->Advertising &
				PHY_AN_AD_100BASETX_HALF) {

				ethcmd.advertising |=
					ADVERTISED_100baseT_Half;
			}
			if (bm_dev->Advertising &
				PHY_AN_AD_100BASETX_FULL) {

				ethcmd.advertising |=
					ADVERTISED_100baseT_Full;
			}
		}
		else {
			ethcmd.autoneg = AUTONEG_DISABLE;
			ethcmd.advertising = 0;
		}

		if(copy_to_user(useraddr, &ethcmd, sizeof(ethcmd)))
			return -EFAULT;
		return 0;
	}
	case ETHTOOL_SSET: {
		if(!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (ethcmd.autoneg == AUTONEG_ENABLE) {
			bm_dev->RequestedLineSpeed = LM_LINE_SPEED_AUTO;
			bm_dev->RequestedDuplexMode = LM_DUPLEX_MODE_UNKNOWN;
			bm_dev->DisableAutoNeg = FALSE;
		}
		else {
			if (ethcmd.speed == SPEED_1000) {
				return -EINVAL;
			}
			else if (ethcmd.speed == SPEED_100) {
				bm_dev->RequestedLineSpeed =
					LM_LINE_SPEED_100MBPS;
			}
			else if (ethcmd.speed == SPEED_10) {
				bm_dev->RequestedLineSpeed =
					LM_LINE_SPEED_10MBPS;
			}
			else {
				return -EINVAL;
			}

			bm_dev->DisableAutoNeg = TRUE;
			if (ethcmd.duplex == DUPLEX_FULL) {
				bm_dev->RequestedDuplexMode =
					LM_DUPLEX_MODE_FULL;
			}
			else {
				bm_dev->RequestedDuplexMode =
					LM_DUPLEX_MODE_HALF;
			}
		}
		BCM4400_PHY_LOCK(pUmDevice, flags);
		b44_LM_SetupPhy(bm_dev);
		BCM4400_PHY_UNLOCK(pUmDevice, flags);
		return 0;
	}
#ifdef ETHTOOL_GWOL
#ifdef BCM_WOL
	case ETHTOOL_GWOL: {
		struct ethtool_wolinfo wol = {ETHTOOL_GWOL};

		wol.supported = WAKE_MAGIC;
		if (bm_dev->WakeUpMode == LM_WAKE_UP_MODE_MAGIC_PACKET)
		{
			wol.wolopts = WAKE_MAGIC;
		}
		else {
			wol.wolopts = 0;
		}
		if (copy_to_user(useraddr, &wol, sizeof(wol)))
			return -EFAULT;
		return 0;
	}
	case ETHTOOL_SWOL: {
		struct ethtool_wolinfo wol;

		if(!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (copy_from_user(&wol, useraddr, sizeof(wol)))
			return -EFAULT;
		
		if ((wol.wolopts & ~WAKE_MAGIC) != 0) {
			return -EINVAL;
		}
		if (wol.wolopts & WAKE_MAGIC) {
			bm_dev->WakeUpMode = LM_WAKE_UP_MODE_MAGIC_PACKET;
		}
		else {
			bm_dev->WakeUpMode = LM_WAKE_UP_MODE_NONE;
		}
		return 0;
        }
#endif
#endif
#ifdef ETHTOOL_GLINK
	case ETHTOOL_GLINK: {
		struct ethtool_value edata = {ETHTOOL_GLINK};

		if (bm_dev->LinkStatus == LM_STATUS_LINK_ACTIVE)
			edata.data =  1;
		else
			edata.data =  0;
		if (copy_to_user(useraddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
	}
#endif
#ifdef ETHTOOL_NWAY_RST
	case ETHTOOL_NWAY_RST: {
		u32 phyctrl;

		if(!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (bm_dev->DisableAutoNeg) {
			return -EINVAL;
		}
		BCM4400_PHY_LOCK(pUmDevice, flags);
		b44_LM_ReadPhy(bm_dev, PHY_CTRL_REG, &phyctrl);
		b44_LM_WritePhy(bm_dev, PHY_CTRL_REG, phyctrl |
			PHY_CTRL_AUTO_NEG_ENABLE |
			PHY_CTRL_RESTART_AUTO_NEG);
		BCM4400_PHY_UNLOCK(pUmDevice, flags);
		return 0;
	}
#endif

	}
	
	return -EOPNOTSUPP;
}
#endif

/* Provide ioctl() calls to examine the MII xcvr state. */
static int bcm4400_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
	u16 *data = (u16 *)&rq->ifr_data;
	u32 value;
	unsigned long flags;

	switch(cmd) {
#ifdef SIOCGMIIPHY
	case SIOCGMIIPHY:
#endif
	case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
		data[0] = bm_dev->PhyAddr;

#ifdef SIOCGMIIREG
	case SIOCGMIIREG:
#endif
	case SIOCDEVPRIVATE+1:		/* Read the specified MII register. */
		BCM4400_PHY_LOCK(pUmDevice, flags);
		b44_LM_ReadPhy(bm_dev, data[1] & 0x1f, (u32 *) &value);
		BCM4400_PHY_UNLOCK(pUmDevice, flags);
		data[3] = value & 0xffff;
		return 0;

#ifdef SIOCSMIIREG
	case SIOCSMIIREG:
#endif
	case SIOCDEVPRIVATE+2:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		BCM4400_PHY_LOCK(pUmDevice, flags);
		b44_LM_WritePhy(bm_dev, data[1] & 0x1f, data[2]);
		BCM4400_PHY_UNLOCK(pUmDevice, flags);
		return 0;

#ifdef SIOCETHTOOL
	case SIOCETHTOOL:
		return netdev_ethtool_ioctl(dev, (void *) rq->ifr_data);
#endif
	default:
		return -EOPNOTSUPP;
	}
	return -EOPNOTSUPP;
}

static void bcm4400_set_rx_mode(struct net_device *dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *)dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
	int i;
	struct dev_mc_list *mclist;

	b44_LM_MulticastClear(bm_dev);
	for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next) {
		b44_LM_MulticastAdd(bm_dev, (u8 *) &mclist->dmi_addr);
	}
	if (dev->flags & IFF_ALLMULTI) {
		if (!(bm_dev->ReceiveMask & LM_ACCEPT_ALL_MULTICAST)) {
			b44_LM_SetReceiveMask(bm_dev,
				bm_dev->ReceiveMask | LM_ACCEPT_ALL_MULTICAST);
		}
	}
	else if (bm_dev->ReceiveMask & LM_ACCEPT_ALL_MULTICAST) {
		b44_LM_SetReceiveMask(bm_dev,
			bm_dev->ReceiveMask & ~LM_ACCEPT_ALL_MULTICAST);
	}
	if (dev->flags & IFF_PROMISC) {
		if (!(bm_dev->ReceiveMask & LM_PROMISCUOUS_MODE)) {
			b44_LM_SetReceiveMask(bm_dev,
				bm_dev->ReceiveMask | LM_PROMISCUOUS_MODE);
		}
	}
	else if (bm_dev->ReceiveMask & LM_PROMISCUOUS_MODE) {
		b44_LM_SetReceiveMask(bm_dev,
			bm_dev->ReceiveMask & ~LM_PROMISCUOUS_MODE);
	}
}

/*
 * Set the hardware MAC address.
 */
static int bcm4400_set_mac_addr(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) dev->priv;

	if (netif_running(dev))
		return -EBUSY;
	memcpy(dev->dev_addr, addr->sa_data,dev->addr_len);
	b44_LM_SetMacAddress(bm_dev, dev->dev_addr);
	return 0;
}



static int bcm4400_suspend (struct pci_dev *pdev, u32 state)
{
	struct net_device *dev = (struct net_device *) pci_get_drvdata(pdev);
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) dev->priv;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;

	if (!netif_running(dev))
		return 0;

	bcm4400_intr_off(pUmDevice);
	netif_carrier_off(dev);
	netif_device_detach (dev);

	/* Disable interrupts, stop Tx and Rx. */
	b44_LM_Halt(bm_dev);
	bcm4400_free_remaining_rx_bufs(pUmDevice);

/*	pci_power_off(pdev, -1);*/
	return 0;
}


static int bcm4400_resume(struct pci_dev *pdev)
{
	struct net_device *dev = (struct net_device *) pci_get_drvdata(pdev);
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) dev->priv;
	struct bm4400_hw * bm_dev = &pUmDevice->lm_dev;

	if (!netif_running(dev))
		return 0;
/*	pci_power_on(pdev);*/
	netif_device_attach(dev);
	b44_LM_InitializeAdapter(bm_dev);
	if (memcmp(dev->dev_addr, bm_dev->NodeAddress, 6)) {
		b44_LM_SetMacAddress(bm_dev, dev->dev_addr);
	}
	atomic_set(&pUmDevice->intr_sem, 0);
	b44_LM_EnableInterrupt(bm_dev);
	return 0;
}


static struct pci_driver bcm4400_pci_driver = {
	name:		bcm4400_driver,
	id_table:	bcm4400_pci_tbl,
	probe:		bcm4400_init_one,
	remove:		__devexit_p(bcm4400_remove_one),
	suspend:	bcm4400_suspend,
	resume:		bcm4400_resume,
};


static int __init bcm4400_init_module (void)
{
	return pci_module_init(&bcm4400_pci_driver);
}


static void __exit bcm4400_cleanup_module (void)
{
	pci_unregister_driver(&bcm4400_pci_driver);
}


module_init(bcm4400_init_module);
module_exit(bcm4400_cleanup_module);

/*
 * Middle Module
 *
 */


int
b44_MM_ReadConfig16(struct bm4400_hw * bm_dev, u32 offset,
	u16 *pValue16)
{
	struct bm4400_dev *pUmDevice;

	pUmDevice = (struct bm4400_dev *) bm_dev;
	pci_read_config_word(pUmDevice->pdev, offset, (u16 *) pValue16);
	return 0;
}

int
b44_MM_ReadConfig32(struct bm4400_hw * bm_dev, u32 offset,
	u32 *pval)
{
	struct bm4400_dev *pUmDevice;

	pUmDevice = (struct bm4400_dev *) bm_dev;
	pci_read_config_dword(pUmDevice->pdev, offset, (u32 *) pval);
	return 0;
}

int
b44_MM_WriteConfig16(struct bm4400_hw * bm_dev, u32 offset,
	u16 Value16)
{
	struct bm4400_dev *pUmDevice;

	pUmDevice = (struct bm4400_dev *) bm_dev;
	pci_write_config_word(pUmDevice->pdev, offset, Value16);
	return 0;
}

int
b44_MM_WriteConfig32(struct bm4400_hw * bm_dev, u32 offset,
	u32 val)
{
	struct bm4400_dev *pUmDevice;

	pUmDevice = (struct bm4400_dev *) bm_dev;
	pci_write_config_dword(pUmDevice->pdev, offset, val);
	return 0;
}

int
b44_MM_AllocateSharedMemory(struct bm4400_hw * bm_dev, u32 BlockSize,
	void * *virt, u32 *phys)
{
	void * pvirt;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	dma_addr_t mapping;

	pvirt = pci_alloc_consistent(pUmDevice->pdev, BlockSize,
					       &mapping);
	if (!pvirt) {
		return -ENOMEM;
	}
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = mapping;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = BlockSize;
	memset(pvirt, 0, BlockSize);
	*virt = (void *) pvirt;
	*phys = (u32) mapping;
	return 0;
}

int
b44_MM_AllocateMemory(struct bm4400_hw * bm_dev, u32 BlockSize,
	void * *virt)
{
	void * pvirt;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;


	/* Maximum in slab.c */
	if (BlockSize > 131072) {
		goto b44_MM_Alloc_error;
	}

	pvirt = kmalloc(BlockSize, GFP_KERNEL);
	if (!pvirt) {
		goto b44_MM_Alloc_error;
	}
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = 0;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = 0;
	/* mem_size_list[i] == 0 indicates that the memory should be freed */
	/* using kfree */
	memset(pvirt, 0, BlockSize);
	*virt = pvirt;
	return 0;

b44_MM_Alloc_error:
	printk(KERN_WARNING "%s: Memory allocation failed - buffer parameters may be set too high\n", pUmDevice->dev->name);
	return -ENOMEM;
}

int
b44_MM_MapMemBase(struct bm4400_hw * bm_dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;

	bm_dev->pMappedMemBase = ioremap_nocache(
		pci_resource_start(pUmDevice->pdev, 0),
			sizeof(bcmenetregs_t) + 128);
	return 0;
}

int
b44_MM_InitializeUmPackets(struct bm4400_hw * bm_dev)
{
	int i;
	struct sk_buff *skb;
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	struct bm4400_packet_buf * buf;
	struct bm4400_packet * packet;

	for (i = 0; i < bm_dev->RxPacketDescCnt; i++) {
		packet = QQ_PopHead(&bm_dev->RxPacketFreeQ.Container);
		buf = (struct bm4400_packet_buf *) packet;
		if (packet == 0) {
			printk(KERN_DEBUG "Bad RxPacketFreeQ\n");
		}
		skb = dev_alloc_skb(packet->u.Rx.RxBufferSize);
		if (skb == 0) {
			buf->skbuff = 0;
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, packet);
			continue;
		}
		buf->skbuff = skb;
		packet->u.Rx.pRxBufferVirt = skb->tail;
		skb->dev = pUmDevice->dev;
		skb_reserve(skb, bm_dev->rxoffset);
		QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
	}
	pUmDevice->rx_buf_repl_thresh = bm_dev->RxPacketDescCnt / 4;
	pUmDevice->rx_buf_repl_panic_thresh = bm_dev->RxPacketDescCnt * 3 / 4;

	return 0;
}

int
b44_MM_GetConfig(struct bm4400_hw * bm_dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	int index = pUmDevice->index;

	if (auto_speed[index] == 0)
		bm_dev->DisableAutoNeg = TRUE;
	else
		bm_dev->DisableAutoNeg = FALSE;

	if (line_speed[index] == 0) {
		bm_dev->RequestedLineSpeed = LM_LINE_SPEED_AUTO;
		bm_dev->DisableAutoNeg = FALSE;
	}
	else {
		if (full_duplex[index]) {
			bm_dev->RequestedDuplexMode = LM_DUPLEX_MODE_FULL;
		}
		else {
			bm_dev->RequestedDuplexMode = LM_DUPLEX_MODE_HALF;
		}

		if (line_speed[index] == 100) {
			bm_dev->RequestedLineSpeed = LM_LINE_SPEED_100MBPS;
		}
		else if (line_speed[index] == 10) {
			bm_dev->RequestedLineSpeed = LM_LINE_SPEED_10MBPS;
		}
		else {
			bm_dev->RequestedLineSpeed = LM_LINE_SPEED_AUTO;
			bm_dev->DisableAutoNeg = FALSE;
			printk(KERN_WARNING "%s: Invalid line_speed parameter (%d), using 0\n", pUmDevice->dev->name, line_speed[index]);
		}

	}
	bm_dev->FlowControlCap = 0;
	if (rx_flow_control[index] != 0) {
		bm_dev->FlowControlCap |= LM_FLOW_CONTROL_RECEIVE_PAUSE;
	}
	if (tx_flow_control[index] != 0) {
		bm_dev->FlowControlCap |= LM_FLOW_CONTROL_TRANSMIT_PAUSE;
	}
	if (auto_flow_control[index] != 0) {
		if (bm_dev->DisableAutoNeg == FALSE) {

			bm_dev->FlowControlCap |= LM_FLOW_CONTROL_AUTO_PAUSE;
			if ((tx_flow_control[index] == 0) &&
				(rx_flow_control[index] == 0)) {

				bm_dev->FlowControlCap |=
					LM_FLOW_CONTROL_TRANSMIT_PAUSE |
					LM_FLOW_CONTROL_RECEIVE_PAUSE;
			}
		}
		else {
			printk(KERN_WARNING "%s: Conflicting auto_flow_control parameter (%d), using 0\n",
				pUmDevice->dev->name, auto_flow_control[index]);
		}

	}

	pUmDevice->timer_interval = HZ / 10;
	pUmDevice->link_interval = HZ / pUmDevice->timer_interval;

	if ((tx_pkt_desc_cnt[index] == 0) ||
		(tx_pkt_desc_cnt[index] > MAX_TX_PACKET_DESC_COUNT)) {

		printk(KERN_WARNING "%s: Invalid tx_pkt_desc_cnt parameter (%d), using %d\n",
			pUmDevice->dev->name, tx_pkt_desc_cnt[index],
			DEFAULT_TX_PACKET_DESC_COUNT);

		tx_pkt_desc_cnt[index] = DEFAULT_TX_PACKET_DESC_COUNT;
	}
	bm_dev->TxPacketDescCnt = tx_pkt_desc_cnt[index];
	if ((rx_pkt_desc_cnt[index] == 0) ||
		(rx_pkt_desc_cnt[index] >= MAX_RX_PACKET_DESC_COUNT)) {

		printk(KERN_WARNING "%s: Invalid rx_pkt_desc_cnt parameter (%d), using %d\n",
			pUmDevice->dev->name, rx_pkt_desc_cnt[index],
			DEFAULT_RX_PACKET_DESC_COUNT);

		rx_pkt_desc_cnt[index] = DEFAULT_RX_PACKET_DESC_COUNT;
	}
	bm_dev->RxPacketDescCnt = rx_pkt_desc_cnt[index];

#ifdef BCM_WOL
	if (enable_wol[index]) {
		bm_dev->WakeUpMode = LM_WAKE_UP_MODE_MAGIC_PACKET;
	}
#endif
	return 0;
}

int
b44_MM_IndicateRxPackets(struct bm4400_hw * bm_dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	struct bm4400_packet * packet;
	struct bm4400_packet_buf * buf;
	struct sk_buff *skb;
	int size;

	while (1) {
		packet = (struct bm4400_packet *)
			QQ_PopHead(&bm_dev->RxPacketReceivedQ.Container);
		if (packet == 0)
			break;
		buf = (struct bm4400_packet_buf *) packet;
#if ! defined(NO_PCI_UNMAP)
		pci_unmap_single(pUmDevice->pdev,
				pci_unmap_addr(buf, map[0]),
				packet->u.Rx.RxBufferSize,
				PCI_DMA_FROMDEVICE);
#endif
		if ((packet->PacketStatus != 0) ||
			((size = packet->PacketSize) > 1518)) {

			/* reuse skb */
#ifdef BCM_TASKLET
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, packet);
#else
			QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
#endif
			pUmDevice->rx_misc_errors++;
			continue;
		}
		skb = buf->skbuff;
		skb_put(skb, size);
		skb->pkt_type = 0;
		skb->protocol = eth_type_trans(skb, skb->dev);
		skb->ip_summed = CHECKSUM_NONE;
		netif_rx(skb);

#ifdef BCM_TASKLET
		buf->skbuff = 0;
		QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, packet);
#else
		skb = dev_alloc_skb(packet->u.Rx.RxBufferSize);
		if (skb == 0) {
			buf->skbuff = 0;
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, packet);
		}
		else {
			buf->skbuff = skb; 
			packet->u.Rx.pRxBufferVirt = skb->tail;
			skb->dev = pUmDevice->dev;
			skb_reserve(skb, bm_dev->rxoffset);
			QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
		}
#endif
	}
	return 0;
}

/* Returns 1 if not all buffers are allocated */
static int
bcm4400_rxfill(struct bm4400_dev * pUmDevice)
{
	struct bm4400_packet * packet;
	struct bm4400_packet_buf * buf;
	struct bm4400_hw * bm_dev = (struct bm4400_hw *) pUmDevice;
	struct sk_buff *skb;
	int queue_rx = 0;
	int ret = 0;

	if (!pUmDevice->opened)
		return ret;

	while ((buf = (struct bm4400_packet_buf *)
		QQ_PopHead(&pUmDevice->rx_out_of_buf_q.Container)) != 0) {
		packet = (struct bm4400_packet *) buf;
		if (buf->skbuff) {
			/* reuse an old skb */
			QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
			queue_rx = 1;
			continue;
		}
		if ((skb = dev_alloc_skb(packet->u.Rx.RxBufferSize)) == 0) {
			QQ_PushHead(&pUmDevice->rx_out_of_buf_q.Container,
				packet);
			ret = 1;
			break;
		}
		buf->skbuff = skb;
		packet->u.Rx.pRxBufferVirt = skb->tail;
		skb->dev = pUmDevice->dev;
		skb_reserve(skb, bm_dev->rxoffset);
		QQ_PushTail(&bm_dev->RxPacketFreeQ.Container, packet);
		queue_rx = 1;
	}
	if (queue_rx) {
		b44_LM_QueueRxPackets(bm_dev);
	}
	return ret;
}

int
b44_MM_IndicateTxPackets(struct bm4400_hw * bm_dev)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	struct bm4400_packet * packet;
	struct bm4400_packet_buf * buf;
	struct sk_buff *skb;
#if ! defined(NO_PCI_UNMAP) && MAX_SKB_FRAGS
	int i;
#endif

	while (1) {
		packet = (struct bm4400_packet *)
			QQ_PopHead(&bm_dev->TxPacketXmittedQ.Container);
		if (packet == 0)
			break;
		buf = (struct bm4400_packet_buf *) packet;
		skb = buf->skbuff;
#if ! defined(NO_PCI_UNMAP)
		pci_unmap_single(pUmDevice->pdev,
				pci_unmap_addr(buf, map[0]),
				pci_unmap_len(buf, map_len[0]),
				PCI_DMA_TODEVICE);
#endif
		dev_kfree_skb_irq(skb);
		buf->skbuff = 0;
		QQ_PushTail(&bm_dev->TxPacketFreeQ.Container, packet);
	}
	if (pUmDevice->tx_full) {
		if (QQ_GetEntryCnt(&bm_dev->TxPacketFreeQ.Container) >=
			(bm_dev->TxPacketDescCnt >> 1)) {

			pUmDevice->tx_full = 0;
			netif_wake_queue(pUmDevice->dev);
		}
	}
	return 0;
}

int
b44_MM_IndicateStatus(struct bm4400_hw * bm_dev, int Status)
{
	struct bm4400_dev * pUmDevice = (struct bm4400_dev *) bm_dev;
	struct net_device *dev = pUmDevice->dev;
	u32 flow_control;

	if (!pUmDevice->opened)
		return 0;

	if (pUmDevice->delayed_link_ind > 0) {
		return 0;
	}
	else {
		if (Status == LM_STATUS_LINK_DOWN) {
			pUmDevice->line_speed = 0;
			netif_carrier_off(dev);
			printk(KERN_ERR "%s: %s NIC Link is Down\n", bcm4400_driver, dev->name);
		}
		else if (Status == LM_STATUS_LINK_ACTIVE) {
			netif_carrier_on(dev);
			printk(KERN_INFO "%s: %s NIC Link is Up, ", bcm4400_driver, dev->name);
		}
	}

	if (Status == LM_STATUS_LINK_ACTIVE) {
		if (bm_dev->LineSpeed == LM_LINE_SPEED_100MBPS)
			pUmDevice->line_speed = 100;
		else if (bm_dev->LineSpeed == LM_LINE_SPEED_10MBPS)
			pUmDevice->line_speed = 10;

		printk("%d Mbps ", pUmDevice->line_speed);

		if (bm_dev->DuplexMode == LM_DUPLEX_MODE_FULL)
			printk("full duplex");
		else
			printk("half duplex");

		flow_control = bm_dev->FlowControl &
			(LM_FLOW_CONTROL_RECEIVE_PAUSE |
			LM_FLOW_CONTROL_TRANSMIT_PAUSE);
		if (flow_control) {
			if (flow_control & LM_FLOW_CONTROL_RECEIVE_PAUSE) {
				printk(", receive ");
				if (flow_control & LM_FLOW_CONTROL_TRANSMIT_PAUSE)
					printk("& transmit ");
			}
			else {
				printk(", transmit ");
			}
			printk("flow control ON");
		}
		printk("\n");
	}
	return 0;
}

int
b44_MM_FreeRxBuffer(struct bm4400_hw * bm_dev, struct bm4400_packet * packet)
{
	struct bm4400_packet_buf * buf;
	struct sk_buff *skb;

	if (packet == 0)
		return 0;
	buf = (struct bm4400_packet_buf *) packet;
	if ((skb = buf->skbuff))
		dev_kfree_skb(skb);
	buf->skbuff = 0;
	return 0;
}


